mysvd <- function (x, n.components = NULL)
{   # Add the percentage of variation explained by n.components components.
    p <- nrow(x)
    n <- ncol(x)
    feature.means <- rowMeans(x)
    x <- t(scale(t(x), center = feature.means, scale = F))
    if (is.null(n.components)) {
        n.components = min(n, p)
    } else n.components <- min(n, p, n.components)
    if (p > n) {
        a <- eigen(t(x) %*% x)
        v <- a$vec[, 1:n.components, drop = FALSE]
        d <- sqrt(a$val[1:n.components, drop = FALSE])
        u <- scale(x %*% v, center = FALSE, scale = d)
        percent <- sum(a$values[1:n.components]) / sum(a$values) * 100
        return(list(u = u, d = d, v = v, feature.means = feature.means, percent = percent))
    }
    else {
        junk <- svd(x)
        nc = min(ncol(junk$u), n.components)
        percent <- sum(junk$d[1:nc]^2) / sum(junk$d^2) * 100
        return(list(u = junk$u[, 1:nc], d = junk$d[1:nc], v = junk$v[,
            1:nc], feature.means = feature.means, percent = percent))
    }
}

impute.check <- function (data) {
# the program is the same as in PAM.R
    colmax <- 0.8
    x <- data
    p <- nrow(x)
    col.nas <- drop(rep(1, p) %*% is.na(x))
    if (any(col.nas > colmax * p)) {
        warning(paste("a column has more than", format(round(colmax *
            100)), "% missing values!"))
        attr(x, "err") <- 1
        return(x)
    }
    attr(x, "err") <- 0
    return(x)
}

predictrisk <- function(xtrain, tme, status, newdata = NULL, ncomp, perc, pca = TRUE,
                        PredictIndexFile = NULL, clintrain = NULL, clintest = NULL) {
    # Note:
    # 1. if (pca == TRUE) cols in xtrain & newdata is obs; eg in gene expression data,
    #       o.w. rows in xtrain & newdata is obs; eg in clinical variables data.
    # 2. newdata may be NULL.
    # 3. If length(perc)=1, 2 groups is considered. If length(perc)=2, 3 groups.
    if (pca) {
       # cur.svd <- superpc:::mysvd(xtrain[ind.genes, -folds[[i]], drop = FALSE], ncomp)
       cur.svd <- mysvd(xtrain, ncomp)
       percent <- cur.svd$percent
       loading <- scale(cur.svd$u, center = FALSE, scale = cur.svd$d)
       xtemp <- t(scale(t(xtrain), center = cur.svd$feature.means, scale = FALSE))
       cur.all <- scale(t(xtemp) %*% cur.svd$u, center = FALSE, scale = cur.svd$d)
       # compute the correlation between pc and significant genes
       gpcor <- cor(t(xtrain), cur.all)
       featuremeans <- cur.svd$feature.means
       if (!is.null(clintrain)) cur.all <- cbind(cur.all, clintrain)
    } else {
       gpcor <- loading <- featuremeans <- NA
       percent <- 100
       cur.all <- xtrain
    }
    pred <- coxcoef <- pmodel <- predindx <- lambda <- lambs <- NA
    # number of columns=ncomp
    options(show.error.messages = FALSE)
    cox <- try(survival::coxph(survival::Surv(tme, status) ~ cur.all))
    options(show.error.messages = TRUE)
    if (!inherits(cox, "try-error")) {
      coxcoef <- cox$coefficients
	  predindx <- apply(cur.all, 1, function(x) sum(x * coxcoef, na.rm=T)) # drop(cur.all %*% coxcoef) suffers from na in coxcoef 2/1/2011
      lambda <- quantile(predindx, perc/100, na.rm = TRUE)
      if (!is.null(newdata)) {
         if (pca) {
            if (nrow(newdata) >= 1) {
              xtemp2 <- t(scale(t(newdata), center = cur.svd$feature.means, scale = FALSE))
              cur.all2 <- scale(t(xtemp2) %*% cur.svd$u, center = FALSE, scale = cur.svd$d)
            } else {
              # still run the analysis using clinical covariates only.
              cur.all2 <- NULL
            }
            if (!is.null(clintest)) cur.all2 <- cbind(cur.all2, clintest)
            lambs <- apply(cur.all2, 1, function(x) sum(x * coxcoef, na.rm=TRUE)) # lambs <- drop(cur.all2 %*% coxcoef)
         } else {
            lambs <- apply(newdata, 1, function(x) sum(x * coxcoef, na.rm=TRUE)) # lambs <- drop(newdata %*% coxcoef)
         }
         if (length(lambda) == 1) {
            pred <- ifelse(lambs <= lambda, "low", "high")
         } else {
            pred <- rep(NA, length(lambs))
            for(i in 1:length(lambs))
               pred[i] <- if (lambs[i] <= lambda[1]) {
                             "low"
                          } else if(lambs[i] <= lambda[2]) {
                             "medium"
                          } else "high"
         }
      }
      # pmodel <- 1-pchisq(2*diff(cox$loglik), ncol(cur.all))
    }
    list(pred = pred, gpcor = gpcor, percent = percent, loading = loading,
         newprogindex = lambs, coxcoef = coxcoef, predindx=predindx,
         featuremeans = featuremeans, threshold = lambda)
}

predictrisklasso <- function(xtrain, tme, status, newdata = NULL, perc,
                        PredictIndexFile = NULL, clintrain = NULL, clintest = NULL, Coefficients) {
    # Note:
    # 1. newdata may be NULL.
    # 2. If length(perc)=1, 2 groups is considered. If length(perc)=2, 3 groups.
    pred <- predindx <- lambda <- lambs <- NA
	if (is.null(clintrain)) {
		predindx <- drop(Coefficients %*% xtrain) # sum(beta * x_i)
	} else {
		predindx <- drop(Coefficients %*% rbind(xtrain, t(clintrain)))
	}
	lambda <- quantile(predindx, perc/100, na.rm = TRUE)
	if (!is.null(newdata)) {
		if (is.null(clintrain)) {
			lambs <- drop(Coefficients %*% newdata)
		} else {
			lambs <- drop(Coefficients %*% rbind(newdata, t(clintest)))
		}
		if (length(lambda) == 1) {
			pred <- ifelse(lambs <= lambda, "low", "high")
		} else {
			pred <- rep(NA, length(lambs))
			for(i in 1:length(lambs))
				pred[i] <- if (lambs[i] <= lambda[1]) {
						 "low"
					  } else if(lambs[i] <= lambda[2]) {
						 "medium"
					  } else "high"
		}
	}
    list(pred = pred, newprogindex = lambs, predindx=predindx, threshold = lambda)
}

predictriskfull <- function(xtrain, xtest, clintrain, clintest, tme, status, ncomp, perc, pca, uselasso,
                            PredictIndexFile, Coefficients = NULL) {
    # called by riskfull().
    out <- list()
	if (!uselasso) {
		predictr <- predictrisk(xtrain, tme, status, xtest, ncomp, perc, pca = pca,
								PredictIndexFile = NULL, clintrain, clintest)
		out$coxcoef <- predictr$coxcoef
	} else {
		predictr <- predictrisklasso(xtrain, tme, status, xtest, perc,
								PredictIndexFile = NULL, clintrain, clintest, Coefficients)
		Coefficients <- Coefficients[1:nrow(xtrain)]
		out$coxcoef <- Coefficients[Coefficients != 0]
	}
    out$predindx <- predictr$predindx
    out$pred <- predictr$pred
    out$loading <- predictr$loading
    out$percent <- predictr$percent
    out$gpcor <- predictr$gpcor
    out$featuremeans <- predictr$featuremeans
    out$threshold <- predictr$threshold
    out$newprogindex <- predictr$newprogindex
    out
}

plotrisk <- function(y, tme, status, filename, main = "", riskgroup = TRUE, cex.main = 1.0) {
#  Plot KM curves.
#  It is possible there is only one value in 'y'. In this case, we define statistic = 0
#  instead of NA and we still make a plot of the KM curve.
   stat <- NA
   if (length(y) > 0) {
      ystring <- sort(unique(as.character(y)))
      y <- factor(y)
      uy <- unique(y)
      #if (length(unique(y)) >= 2) {
        stat <- if (length(ystring) > 1) survival::survdiff(survival::Surv(tme, status) ~ y)$chisq else 0
        if (!is.null(filename)) {
            png(filename)
            par(mar=c(5,4,4,2)+.1)
            if (length(uy) == 2 & riskgroup) {
               col <- c("red", "blue")
               legend.text <- c("High risk", "Low risk")
               lty <- 1:2
            } else if (length(uy) == 3 & riskgroup){
               col <- c("red", "green", "blue")
               # strata is sorted by alphabet.
               legend.text <- c("High risk", "Low risk", "Medium risk")
               lty <- 1:3
            } else if (length(uy) == 1 & riskgroup) {
               # added 9/26/06
               col <- ifelse (uy == "high", "red", "blue")
               legend.text <- ifelse(uy == "high", "High risk", "Low risk")
               lty <- 1
            } else if (!riskgroup) {
               legend.text <- ystring
               col <- rainbow(length(ystring))
               lty <- 1:length(ystring)
            }
            plot(survival::survfit(survival::Surv(tme, status) ~ y), conf.int = FALSE,
                 lty = lty, col = col,
                 xlab = "time", ylab = "Survival function")
            legend("bottomleft", legend=legend.text, col=col, lty=lty, bty="n")
            title(main = main, cex.main = cex.main)
            dev.off()
        }
      #}
   }
   stat
}


riskfull <- function(xtrain, tme, status, clintrain, ncomp, alpha, perc, xtest = NULL, clintest = NULL,
                     PredictIndexFile = NULL, PredictIndexclin = NULL, PredictIndexFilecombo = NULL,
                     out.CC = NULL, uselasso, mixing, pcrgenes) {
    # Run risk group prediction using full dataset (xtrain).
    # simplified from riskcv() function.
    # Input:
    #   out.CC: an object returned from CheckCovariate() function.
	#   uselasso: Fit the cox regression by using PC or lasso method.
    # Output:
    #   modelE : list of many components, fit gene expression only,
    #            pred,coxcoef,predindx,percent,gpcor,loading.
    #   modelCl: list of many componnets, fit covariates only.
    #            pred,coxcoef,predindx.
    #   modelCo: list of many components, fit gene expression and covariates.
    #            predi,coxcoef,predindx,percent,gpcor,loading.
    # Note:
    #   $pred: if no genes were found, it is not defined (NULL if we want to extract its value from a list).
    #          if no predict samples were given, it is defined as NA.
    ngenes <- nrow(xtrain)
    trndf <- data.frame(tme=tme, status = status)
	lambda <- NULL # optimal lambda. Used if uselasso=TRUE
    # fit gene by gene gene-expression to the coxph model.
    imodel <- 1;  nvar <- 1
    if (uselasso) {
		fit <- glmnet::glmnet(t(xtrain), survival::Surv(tme, status), family="cox", maxit = 1000, alpha = mixing)
		nonzerogenes <- sapply(fit$lambda, function(x) sum(drop(coef(fit, s=x)) != 0)) # number of nonzero-coef genes for each lambda
		optlambda <- fit$lambda[which.min(abs(nonzerogenes - pcrgenes))]
		Coefficients <- as.vector(drop(coef(fit, s = optlambda))) # Ting forced it to vector 9/8/2017
		ind.genes <- which(Coefficients != 0)
		if (length(ind.genes) >= 1) {
			# Do not need to fit the model again. Do not need to subset variables.
			modelE <- predictriskfull(xtrain,
									xtest,
									NULL, NULL, tme, status, ncomp, perc,
									pca = ifelse(uselasso, FALSE, TRUE), uselasso, PredictIndexFile, Coefficients)
			modelE$lambda <- optlambda
			lambda <- optlambda
		} else {
			modelE <- list(ind.genes=NULL)
			return(invisible(list(modelE=modelE, lambda=lambda)))
		}
	} else {
	    ind <- apply(trndf, 1, function(x) all(is.finite(x)))
		ind <- seq(1, length(tme))[ind & apply(xtrain, 2, function(x) all(is.finite(x)))]
		# sort the data according to the time order.
		sorted <- order(trndf$tme[ind])
		trndf <- trndf[ind[sorted], ]; trnx <- xtrain[, ind[sorted]]
		gc()
		coxF <- try(.Fortran(getNativeSymbolInfo("coxfitc"), nrow(trndf), as.integer(ngenes), as.integer(0),
						  as.double(trndf$tme), as.integer(trndf$status), t(trnx), as.double(0), as.integer(imodel),
						  double(2*nrow(trndf)+2*nvar*nvar+3*nvar), logdiff = double(ngenes)))
		gc()
		if (!inherits(coxF, "try-error")) {
		   pval <- 1 - pchisq(coxF$logdiff, 1)
		   ind.genes <- which(pval <= alpha)
		} else stop("Error in calling coxph in riskfull()")
		if (length(ind.genes) >= 1) {
			modelE <- predictriskfull(xtrain[ind.genes, , drop = FALSE],
									xtest[ind.genes, , drop = FALSE],
									NULL, NULL, tme, status, ncomp, perc,
									pca = ifelse(uselasso, FALSE, TRUE), uselasso, PredictIndexFile)
		} else {
			modelE <- list(ind.genes=NULL)
			return(invisible(list(modelE=modelE, lambda=lambda)))
		}
		modelE$indexconst <- modelE$featuremeans %*% modelE$loading %*% modelE$coxcoef
		modelE$pval <- pval[ind.genes]
	}
	modelE$ind.genes <- ind.genes

    modelCl <- list()
    modelCo <- list()
    if (!is.null(clintrain)) {
       # fit solely the clinical variables to the coxph model.
       # no rotation is needed in clintrain & clintest.
       if (out.CC$nlevel > 4) {
          modelCl <- predictriskfull(clintrain, clintest,
                               NULL, NULL,
                               tme, status, ncomp, perc, pca = FALSE, uselasso =FALSE, PredictIndexclin)
       }  else modelCl <- list()
       # fit both the clinical variables and the gene expression.
       imodel <- 2; nvar <- ncol(clintrain) + 1
	   if (uselasso) {
			nclintrain <- ncol(clintrain)
			xtrainclin <- as.matrix(cbind(t(xtrain), clintrain))
			# Possible error in glmnet():
			# Error in validObject(.Object) :
            #   invalid class "dgCMatrix" object: length(Dimnames[[2]])' must match Dim[2]
            # In addition: Warning message:
            # from glmnet Fortran code - Outer loop convergence not reached after 1000 iterations
            fit3 <- try(glmnet::glmnet(xtrainclin, survival::Surv(tme, status), penalty.factor = c(rep(1, ngenes), rep(0, nclintrain)),
                                       family="cox", maxit = 1000, alpha = mixing))
            if (!inherits(fit3, "try-error")) {
				nonzerogenes <- sapply(fit3$lambda, function(x) sum(drop(coef(fit3, s=x)) != 0)) - nclintrain #Ming-Chung 04/23/15 excluded the clinical covariates when it is counting the number of non-zero coefficients
				optlambda <- fit3$lambda[which.min(abs(nonzerogenes - pcrgenes))]
				Coefficients <- as.vector(drop(coef(fit3, s = optlambda))) # Ting forced it to vector 9/8/2017
				ind.genes <- which(Coefficients != 0)
				ind.genes <- ind.genes[ind.genes <= nrow(xtrain)] # exclude clinical covariates
				if (length(ind.genes) >= 1) { # does not change as riskcv()
					modelCo <- predictriskfull(xtrain,
											xtest,
											clintrain, clintest,
											tme, status, ncomp, perc, pca = TRUE, uselasso, PredictIndexFilecombo, Coefficients)
					modelCo$lambda <- optlambda
					lambda <- c(lambda, optlambda)
				} else {
					modelCo <- list(ind.genes=NULL)
					lambda <- c(lambda, optlambda) #Ting 04/16/15 to fix a bug when there are no significant genes found
					return(invisible(list(modelE = modelE, modelCl = modelCl, modelCo = modelCo, lambda = lambda)))
				}
			} else {
				modelCo <- list(ind.genes=NULL)
				return(invisible(list(modelE = modelE, modelCl = modelCl, modelCo = modelCo, lambda = lambda)))
			}
		} else {
			clintrain2 <- clintrain[ind[sorted],]
			gc()
			coxF <- try(.Fortran(getNativeSymbolInfo("coxfitc"), nrow(trndf), as.integer(ngenes), ncol(clintrain),
					   as.double(trndf$tme), as.integer(trndf$status), t(trnx),
					   as.double(clintrain2), as.integer(imodel),
					   double(2*nrow(trndf)+2*nvar*nvar+3*nvar), logdiff = double(ngenes)))
			gc()
			if (!inherits(coxF, "try-error")) {
				pval <- 1 - pchisq(coxF$logdiff, 1)
				ind.genes <- which(pval <= alpha)
				if (length(ind.genes) >= 1) { # does not change as riskcv()
					modelCo <- predictriskfull(xtrain[ind.genes, ,drop = FALSE],
											xtest[ind.genes, , drop = FALSE],
											clintrain, clintest,
											tme, status, ncomp, perc, pca = TRUE, uselasso, PredictIndexFilecombo)
				} else {
					modelCo <- list(ind.genes=NULL)
					return(invisible(list(modelE = modelE, modelCl = modelCl, modelCo = modelCo)))
				}
			} else stop("Error in calling coxph in combo data")
			modelCo$pval <- pval[ind.genes]
			modelCo$indexconst <- modelCo$featuremeans %*% modelCo$loading %*% modelCo$coxcoef[1:ncomp]
		}
		modelCo$ind.genes <- ind.genes
    }
    invisible(list(modelE = modelE, modelCl = modelCl, modelCo = modelCo, lambda = lambda))
}

CheckCovariate <- function(clin) {
    # Compute the total number of levels in all combination of categorical covariates
    # Input:
    #    clin: covariate matrix
    # Output:
    #    nlevel:the levels of all categorical covariates. If at least one of the
    #           covariate is continuous, nlevel = 5 (4+1 where 4 is the boundary).
    #    levelname: level name. Each one corresponds to one combination of levels from one row.
    clin <- data.frame(clin)
    for(i in 1:ncol(clin)) if (!is.factor(clin[,i])) return(list(nlevel=5))
    #levelunique <- unique.matrix(clin)
    #levelunique <- sort(apply(levelunique, 1, paste, collapse = " "))
    levelname <- apply(clin, 1, paste, collapse = "-")
    nlevel <- length(unique(levelname))
    list(nlevel = nlevel, levelname = levelname)
}

riskcv <- function(xtrain, tme, status, clin, folds, ncomp, alpha, perc,
                   PlotFileName = NULL, PlotFileNameclin = NULL, PlotFileNamecomb = NULL,
                   observed = FALSE, clin.Orig, uselasso, mixing, lambda, WorkPath) {
#  Input:
#     xtrain: gene x array expression matrix
#     clin: array x var matrix
#     clin.Orig: the original covarites (before creating dummy variables if there are discrete
#                covariate.) It is used to check the number of levels in the training set
#                and each CV training set.
#     lambda: lambda parameter used in glmnet().
#  Output:
#     stat3: vector of length 3, chisquare statistic for 2 groups.
#     CVsupport: (ngenes x 2) CV support, only if observed = TRUE.
#  Note:
#     For the covariate alone model, we will not do the permutation even permutation test is requested.
#     We will only report the parametric p-value.
#     If PlotFileName=NULL, KM plot will not be created.
   errmsg <- ""
   ngenes <- nrow(xtrain)
   y <- y.clin <- y.comb <- character(length(tme))
   # y: the predict risk group by fitting PCA to the genes selected by univariate coxph model
   # y.clin: the predict risk group by fitting clinical variables coxph model
   # y.comb: the preidct risk group by fitting PCA to the genes selected by comb coxph models.
   if (observed) {
     CV <- matrix(0, nr = ngenes, nc = 2)

     colnames(CV) <- c("modelE", "modelCo")
   }
   newprogindex <- matrix(NA, nr = length(tme), nc=3)
   colnames(newprogindex) <- c("modelE", "modelCl", "modelCo")
   if (!is.null(clin))   out.CC <- CheckCovariate(clin.Orig)
   for(i in 1:length(folds)) {
		# handle missing values in status, tme or training.
		trndf <- data.frame(status = status[-folds[[i]]], tme = tme[-folds[[i]]])
		trnx <- xtrain[,-folds[[i]], drop = FALSE]
		ind <- apply(trndf, 1, function(x) all(is.finite(x)))
		ind <- seq(1, nrow(trndf))[ind & apply(trnx, 2, function(x) all(is.finite(x)))]
		# sort the data according to the time order.
		sorted <- order(trndf$tme[ind])
		trndf <- trndf[ind[sorted], ]; trnx <- trnx[, ind[sorted], drop=FALSE]
		if (observed) {
			flush.console()
			if (i%%10 != 0) {
				# cat(i," ", file=stderr())
				cat(i," ")
			} else {
				# cat(i,"\n", file=stderr())
				cat(i,"\n")
			}
		}
		imodel <- 1;  nvar <- 1
		if (uselasso) {
			fit <- glmnet::glmnet(t(trnx), survival::Surv(trndf$tme, trndf$status), family="cox", maxit = 1000, alpha = mixing)
			Coefficients <- try(as.vector(drop(coef(fit, s = lambda[1])))) # Ting forced it to vector 9/8/2017
			if (inherits(Coefficients, "try-error")) return(list(errmsg=geterrmessage()))
			ind.genes <- which(Coefficients != 0)
		} else {
			# fit gene by gene gene-expression to the coxph model, and select genes
			# and obtain predicted groups for left-out samples.
			gc()
			coxF <- try(.Fortran(getNativeSymbolInfo("coxfitc"), nrow(trndf), as.integer(ngenes), as.integer(0),
							as.double(trndf$tme), as.integer(trndf$status), t(trnx), as.double(0), as.integer(imodel),
							double(2*nrow(trndf)+2*nvar*nvar+3*nvar), logdiff = double(ngenes)))
			gc()
			if (!inherits(coxF, "try-error")) {
				pval <- 1 - pchisq(coxF$logdiff, 1)
				ind.genes <- which(pval <= alpha)
				if (length(ind.genes) == 0) {
					ind.genes <- which.min(pval)
				}
			} else stop("Error in calling coxph in riskcv")
		}
		if (length(ind.genes) >= 1) {
			if (!uselasso) {
				probj <- predictrisk(xtrain[ind.genes, -folds[[i]], drop = FALSE],
                              tme[-folds[[i]]], status[-folds[[i]]],
                              xtrain[ind.genes,folds[[i]], drop = FALSE], ncomp, perc)
			} else {
				probj <- predictrisklasso(xtrain[, -folds[[i]], drop = FALSE],
								tme[-folds[[i]]], status[-folds[[i]]],
								xtrain[, folds[[i]], drop = FALSE], perc, Coefficients = Coefficients)
			}
			y[folds[[i]]] <- probj$pred
			newprogindex[folds[[i]], 1] <- probj$newprogindex
			if (observed) {
				CV[ind.genes, 1] <- CV[ind.genes, 1] + 1
			}
		}

		if (!is.null(clin)) {
			# fit solely the clinical variables to the coxph model and obtain predicted groups
			# for left-out samples.
			# Note:
			#     1. if the number of levels <=4 in the training set, do not run prediction in CV.
			#     2. if the number of levels <=4 in the CV's training set, do not run prediction.
			clin1 <- clin[-folds[[i]],,drop = FALSE]
			clin2 <- clin1[ind[sorted],,drop = FALSE] # needed in model III.
			if (observed & out.CC$nlevel > 4) {
				out.CC2 <- CheckCovariate(clin.Orig[-folds[[i]],,drop = FALSE])
				if (out.CC2$nlevel > 4) {
					options(show.error.messages = FALSE)
					cox <- try(survival::coxph(survival::Surv(trndf$tme, trndf$status) ~ clin2))
					options(show.error.messages = TRUE)
					if (!inherits(cox, "try-error")) {
						#pvalclin <- 2*(1 - pnorm(abs(coef(cox) / sqrt(diag(cox$var)))))
						probj <- predictrisk(clin1,
                                        tme[-folds[[i]]], status[-folds[[i]]],
                                        clin[folds[[i]], ,drop = FALSE], ncomp, perc, pca = FALSE)
						y.clin[folds[[i]]] <- probj$pred
						newprogindex[folds[[i]], 2] <- probj$newprogindex
					} # else stop("Error in calling coxph in fitting covariate data.")
				}
			}
			# fit two models and do a LRT to select genes
			# 1) the clinical variables 2) the gene express data and clinical variables to
			# the coxph models. Finally obtain the predicted group.
			imodel <- 2; nvar <- ncol(clin) + 1
			if (uselasso) {
				nclintrain <- ncol(clin2)
				xtrainclin <- as.matrix(cbind(t(trnx), clin2))
				fit3 <- try(glmnet::glmnet(xtrainclin, survival::Surv(trndf$tme, trndf$status), penalty.factor = c(rep(1, ngenes), rep(0, nclintrain)),
								family="cox", maxit = 1000, alpha = mixing))
                if (!inherits(fit3, "try-error")) {
				   Coefficients <- as.vector(drop(coef(fit3, s = lambda[2]))) # Ting forced it to vector 9/8/2017
				   ind.genes <- which(Coefficients != 0)
				   ind.genes.comb <- ind.genes[ind.genes <= nrow(xtrain)] # exclude clinical covariates
				} else {
				   ind.genes.comb <- c()
				}
			} else {
				gc()
				coxF <- try(.Fortran(getNativeSymbolInfo("coxfitc"), nrow(trndf), as.integer(ngenes), ncol(clin),
							  as.double(trndf$tme), as.integer(trndf$status), t(trnx), as.double(clin2), as.integer(imodel),
							  double(2*nrow(trndf)+2*nvar*nvar+3*nvar), logdiff = double(ngenes)))
				gc()
				if (!inherits(coxF, "try-error")) {
					pvalcomb <- 1 - pchisq(coxF$logdiff, 1)
					ind.genes.comb <- which(pvalcomb <= alpha)
					if (length(ind.genes.comb) == 0) {
						# based on Rich's instruction on Oct 18 2007
						# Pick the most significant gene and combine with the covariates
						ind.genes.comb <- which.min(pvalcomb)
					}
				} else stop("Error in calling coxph in CV")
			}
			if (length(ind.genes.comb) >= 1) {
				if (!uselasso) {
					probj <- predictrisk(xtrain[ind.genes.comb, -folds[[i]], drop = FALSE],
									  tme[-folds[[i]]], status[-folds[[i]]],
									  xtrain[ind.genes.comb,folds[[i]], drop = FALSE], ncomp, perc,
									  clintrain=clin[-folds[[i]],,drop=FALSE],
									  clintest=clin[folds[[i]],,drop=FALSE])
				} else {
					probj <- predictrisklasso(xtrain[, -folds[[i]], drop = FALSE],
										tme[-folds[[i]]], status[-folds[[i]]],
										xtrain[, folds[[i]], drop = FALSE], perc,
										clintrain=clin[-folds[[i]],,drop=FALSE],
										clintest=clin[folds[[i]],,drop=FALSE], Coefficients = Coefficients)
				}
				y.comb[folds[[i]]] <- probj$pred
				newprogindex[folds[[i]], 3] <- probj$newprogindex
				if (observed) {
					CV[ind.genes.comb, 2] <- CV[ind.genes.comb, 2] + 1
				}
			}
		} # end of if (!is.null(clin))
	} # end of for(i in 1:length(folds))
	if (observed) {
		# cat("\n", file = stderr())
		# cat("\n")
		CV <- CV / length(folds) * 100
	}
	ind <- which(y != "") # exclude nlevel <=4 or problem with coxph().
	# It is possible 1. all(ind == "").
	#                2. all predicted risk groups has just one value.
	stat <- plotrisk(y[ind], tme[ind], status[ind], PlotFileName,
                    main = "Fitting by gene expression only")
	y_plotKM <- y[ind]
	tme_plotKM <- tme[ind]
	status_plotKM <- status[ind]
	save(y_plotKM, tme_plotKM, status_plotKM, file = paste(WorkPath, "/plotKM.rda", sep=""))
	if (!is.null(clin)) {
		if (observed & out.CC$nlevel > 4) {
			ind <- which(y.clin != "")
			stat.clin <- plotrisk(y.clin[ind], tme[ind], status[ind], PlotFileNameclin,
									main = "Fitting by covariates only")
			y_plotKM <- y.clin[ind]
			tme_plotKM <- tme[ind]
			status_plotKM <- status[ind]
			save(y_plotKM, tme_plotKM, status_plotKM, file = paste(WorkPath, "/plotKMclin.rda", sep=""))

		} else if (observed & out.CC$nlevel <= 4) {
			stat.clin <- plotrisk(out.CC$levelname, tme, status, PlotFileNameclin, main = "", riskgroup = FALSE)
			y_plotKM <- out.CC$levelname
			tme_plotKM <- tme
			status_plotKM <- status
			save(y_plotKM, tme_plotKM, status_plotKM, file = paste(WorkPath, "/plotKMclin.rda", sep=""))
		} else {
			# no permutation test in the covariate alone model 8/30/06.
			stat.clin <- NA
		}
		ind <- which(y.comb != "")
		stat.comb <- plotrisk(y.comb[ind], tme[ind], status[ind], PlotFileNamecomb,
								main = "Fitting by both gene expression\nand covariates")
		y_plotKM <- y.comb[ind]
		tme_plotKM <- tme[ind]
		status_plotKM <- status[ind]
		save(y_plotKM, tme_plotKM, status_plotKM, file = paste(WorkPath, "/plotKMcomb.rda", sep=""))

	} else {
		stat.clin <- stat.comb <- NA
	}
	stat3 <- c(stat, stat.clin, stat.comb)
	if (observed) {
		pred <- cbind(y, y.clin, y.comb)
		invisible(list(stat3 = stat3, CVsupport = CV, pred = pred, newprogindex = newprogindex, errmsg=errmsg))
	} else {
		invisible(list(stat3 = stat3, newprogindex = newprogindex, errmsg=errmsg))
	}
}

survRiskMain <- function(status, tme, training, arr1, nperm, nfold, clin, perc,
                     alpha, ncomp, uselasso, mixing, pcrgenes, landmarktime,
                     PlotKMFileName, PlotKMFileNameclin, PlotKMFileNamecomb,
                     PredictIndexFile, PredictIndexFileclin, PredictIndexFilecomb,
                     PlotrocFileName, WorkPath) {
  # This function is called foo() in BRB-ArrayTools
  # Combine the R statements to a function so when errors happen, they can be handled smoothly.
  # err = 0, no error
  # err = 1, # of levels < 2 for some categorical covariate.
  # err = 2, 'impute', 'glmnet' or 'survivalROC' package can not be installed.
  # err = 3, imputation fails because one column has more than 80% missing values.
  # err = 4, some genes have missing values on all arrays.
  # err = 5, model can not be estimated.
  # err = 6, no genes were found at the alpha level.
	# err = 7, The glmnet function experiences a problem.
	# err = 8, status is not 0/1 values.

  nsamp <- sum(training)
  # ntest <- dim(arr1)[2] - nsamp
  # ngenes <- nrow(arr1)

  err <- 0
	errmsg <- ""
  # Check if some covaraites are categorical but have only 1 level
  if (!is.null(clin)) {
    for(i in 1:ncol(clin))
      if (is.factor(clin[,i]) & nlevels(clin[,i])<2) {
         #winDialog("ok", "The number of levels for some categorical covariates is less than two. Please correct it and run the analysis again.")
         return(list(err = 1))
      }
  }
	if (is.data.frame(arr1)) arr1 <- data.matrix(arr1)
  if(any(is.na(arr1))) {
#        if (!any(installed.packages()[,1] == "impute")) {
# 	       PKGNAME <- "impute"
# 		     IsBioC <- TRUE
# 		     if (!LoadRPackage(PKGNAME, isBioconductor=IsBioC)) {
# 				   require(tcltk)
# 				   tkmessageBox(message=paste(PKGNAME, "cannot be downloaded or installed."),
# 									icon="info",type="ok")
# 				   return(list(err == 2))
# 	       }
#        }
#        if (!require(impute)) {
#           return(list(err = 2))
#        }
     print("Imputing missing values.")
     IsImpute <- TRUE
     if (attr(impute.check(arr1), "err") == 1) {
         #winDialog(type="ok",
         #          message=paste("KNN imputation stops because at least one column ",
         #                        "of data has more than 80% missing values. ",
         #                        "Make sure no arrays have more than 80% missing values ",
         #                        "and rerun the analysis.", sep = ""))
         return(list(err = 3))
     }
     arr1 <- impute::impute.knn(arr1, k = min(nrow(arr1)-1, 10))
     if (!is.matrix(arr1)) arr1 <- arr1$data
     if (any(is.na(arr1))) {
         #winDialog(type="ok",
         #          message=paste("Missing data was still found after the imputation. It may be some genes have missing values on all arrays. ",
         #                        "Please subset genes and rerun the analysis.", sep = ""))
         return(list(err = 4))
     }
  } else {
    IsImpute <- FALSE
  }
# 	if (uselasso) {
#        if (!any(installed.packages()[,1] == "glmnet")) {
#           PKGNAME <- "glmnet"
#   	      IsBioC <- FALSE
# 		      if (!LoadRPackage(PKGNAME, isBioconductor=IsBioC)) {
# 				    require(tcltk)
# 				    tkmessageBox(message=paste(PKGNAME, "cannot be downloaded or installed."),
# 									icon="info",type="ok")
# 				    return(list(err == 2))
# 	        }
#        }
#        if (!require(glmnet)) {
# 		   return(list(err = 2))
# 		}
#	}
#   if (!any(installed.packages()[,1] == "survivalROC")) {
#     PKGNAME <- "survivalROC"
#     IsBioC <- FALSE
#     if (!LoadRPackage(PKGNAME, isBioconductor=IsBioC)) {
# 	    require(tcltk)
# 	    tkmessageBox(message=paste(PKGNAME, "cannot be downloaded or installed."),
# 						icon="info",type="ok")
# 	    return(list(err = 2))
#     }
#   }
# 	if (!require(survivalROC))  return(list(err = 2))

  xtrain <- arr1[,training == 1, drop =FALSE]
  tmetrain <- tme[training == 1]
  statustrain <- status[training == 1]
  if (!is.null(clin)) {
    clin.Orig <- clin # contain training and predict samples
    clin <- model.matrix(lm(rnorm(nrow(clin)) ~ .,clin))[, -1, drop = FALSE]
  }
  clintrain <- if (!is.null(clin)) clin[training == 1,,drop = FALSE] else NULL
  # the subset covariates may have singular covariate. So we need to check this condition first.
  if (!is.null(clin)) {
     clinalias <- alias(lm(rnorm(nrow(clintrain)) ~ clintrain))$Complete
     if (!is.null(clinalias)) {
        print(clinalias)
        #winDialog("ok",
        #paste("The model is not estimable with the covariates in the training set.",
        #      "The analysis is aborted now.",
        #      "Please double check the covariates in the training set and rerun the analysis again!"))
        return(list(err=5))
     }
  }
  set.seed(1234) # used in 2 places
  if (nfold == nsamp) {
     folds <- split(1:nsamp, 1:nsamp)
  } else if (nfold == 10) {
     folds <- split(sample(1:nsamp), rep(1:nfold, length = nsamp))
  } else {
    stop("Are you sure the nfold is right?")
  }
  if (sum(training ==0) > 0) {
     xtest <- arr1[, training == 0, drop = FALSE]
     clintest <- clin[training == 0, , drop = FALSE]
  } else xtest <- clintest <- NULL
  if (!is.null(clin)) {
    out.CC <- CheckCovariate(clin.Orig[training == 1,,drop = FALSE])
    cvrt <- clin.Orig[training == 1,,drop = FALSE]
  } else {
    out.CC <- cvrt <- NULL
  }
  # cat("Starting the analysis using the full training data.\n", file=stderr())
  cat("Starting the analysis using the full training data.\n")
  gc()
	labpred <- riskfull(xtrain, tmetrain, statustrain, clintrain, ncomp, alpha, perc, xtest,
					clintest, PredictIndexFile, PredictIndexFileclin, PredictIndexFilecomb,
					out.CC = out.CC, uselasso=uselasso, mixing=mixing, pcrgenes=pcrgenes)
  if (length(labpred$modelE$ind.genes) == 0) {
     # The condition of the minimal alpha level is based on the genes obtained in CV rather than
     # the complete training data. This condition may not be the same as in the survival analysis.
     #winDialog(type="ok",
     #          message=paste("No significant genes were found at the selected ",
     #                        "alpha level. Please increase the alpha value and ",
     #                        "run the analysis again.", sep = ""))
     return(list(err = 6))
  }
  # cat("Cross-validating the model.\n", file=stderr())
  cat("Cross-validating the model.\n")
  # cat("The number of folds in CV is ", length(folds), "\n", file=stderr())
  cat("The number of folds in CV is ", length(folds), "\n")
  # options(show.error.messages = FALSE, warn = -1)
  elptme <- system.time(statobs <- riskcv(xtrain, tmetrain, statustrain, clintrain, folds, ncomp,
                                          alpha, perc, PlotKMFileName, PlotKMFileNameclin,
                                          PlotKMFileNamecomb, observed = TRUE, clin.Orig = cvrt,
										uselasso=uselasso, mixing=mixing, lambda=labpred$lambda, WorkPath))[3] #9/14
	# Compute AUC(t) and give time-dependent ROC curve for the observed data
  aucobs <- rep(NA, 3)
  mdlroc <- seq(1,3)[sapply(1:3, function(i) all(is.finite(statobs$newprogindex[, i])))]
	plotted <- FALSE
	if (length(mdlroc) > 0) {
		png(PlotrocFileName)
		for(i in mdlroc) {
		  rocout <- try(survivalROC::survivalROC(Stime=tmetrain,  status=statustrain,  marker = statobs$newprogindex[, i],
									predict.time = landmarktime, span = 0.25*nsamp^(-0.20) ))
		  if (!inherits(rocout, "try-error")) {
			 if (plotted) par(new = TRUE)
			 aucobs[i] <- rocout$AUC
			 plot(rocout$FP, rocout$TP, type="l", xlim=c(0,1), ylim=c(0,1),  lty=i, xlab="", ylab="",
				  main =paste("Landmark time:", landmarktime))
			 plotted <- TRUE
		  }
		}
		abline(0, 1, col="gray")
		aucval <- paste("AUC=", paste(round(aucobs[mdlroc], 3), c("(genes)", "(covariates)", "(combined)")[mdlroc], collapse=",", sep=""))
		title(xlab=paste( "1 - specificity", "\n", aucval), ylab="sensitivity")
		legend("topleft", lty=mdlroc, c("genes", "covariates", "combined")[mdlroc])
		dev.off()

		save(statobs, tmetrain, statustrain, landmarktime, nsamp, file = paste(WorkPath, "/plotROCCurveModelE.rda", sep=""))
		# AUC[3] records the difference of AUC between covariates only and combined model.
		# This will be used in the permutation test.
		aucobs[3] <- aucobs[2] - aucobs[1]
	}
	if (nperm > 0) {
     statprm <- matrix(NA, nr = nperm, nc = 3)
	   aucprm <- matrix(NA, nr= nperm, nc=3)
     print("Starting permutation test.")
     # cat("Approximately ", nperm*elptme, "seconds is required to run the permutation test\n", file=stderr())
     cat("Approximately ", nperm*elptme, "seconds is required to run the permutation test\n")
     # cat("(", nperm, " permutations)\n", sep = "", file =stderr())
     cat("(", nperm, " permutations)\n", sep = "")
	   iset <- if(!is.null(clin))  c(1,3) else  1
     for(b in 1:nperm) {
        flush.console()
        if (b%%10 != 0) {
          # cat(b," ", file = stderr())
          cat(b," ")
        } else {
          # cat(b,"\n", file = stderr())
          cat(b,"\n")
        }
        perm <- sample(1:nsamp)
        if (!is.null(clin)) {
          clinp <- clintrain[perm,,drop = FALSE]
          cvrtp <- cvrt[perm,,drop = FALSE]
        } else {
           clinp <- cvrtp <- NULL
        }
        tmp <- riskcv(xtrain, tmetrain[perm], statustrain[perm], clinp, folds, ncomp,
                               alpha, perc, observed = FALSE, clin.Orig = cvrtp,
							   PlotFileName = NULL, PlotFileNameclin = NULL, PlotFileNamecomb = NULL,
							 uselasso = uselasso, mixing = mixing, lambda = labpred$lambda, WorkPath)
        if (plotted) {
		      if (tmp$errmsg == "") {
			      statprm[b, ] <- tmp$stat3
			      # Comupute AUC(t) for the permuted data
    			  for(i in iset) {
    				aucprm[b, i] <- survivalROC::survivalROC(Stime=tmetrain[perm],  status=status[perm],  marker = tmp$newprogindex[, i],
    									predict.time = landmarktime, span = 0.25*nsamp^(-0.20) )$AUC
    			  }
    			  # Clinical covariates has no permutation!
    			  aucprm[b, 3] <- aucobs[2] - aucprm[b, 3]
		      } else {
			      return(list(err=7, errmsg = tmp$errmsg))
		      }
	      } # end of if (plotted)
      } # end of for (b in 1:nperm)
      cat("\n")
      # cat("\n", file = stderr())
      # Permutation based pvalue for log rank test for model 1 and 3
      # parametric chi-square log rank test for model 2.
      pval <- rep(NA, 3)
      pval[1] <- sum(statprm[,1] >= statobs$stat3[1], na.rm = TRUE) / sum(!is.na(statprm[,1]))
      pval[2] <- 1 - pchisq(statobs$stat3[2], length(perc))
      pval[3] <- sum(statprm[,3] >= statobs$stat3[3], na.rm = TRUE) / sum(!is.na(statprm[,3]))
	    pval.roc <- rep(NA, 3)
	    if (plotted) {
  		  pval.roc[1] <- sum(aucprm[, 1] >= aucobs[1], na.rm= TRUE) / sum(is.finite(aucprm[,1]))
	  	  # pval.roc[2] <- sum(aucprm[, 2] >= aucobs[2], na.rm= TRUE) / sum(is.finite(aucprm[,2]))
		    pval.roc[3] <- sum(aucprm[, 3] >= aucobs[3], na.rm= TRUE) / sum(is.finite(aucprm[,3]))
	    }
      # cat("Finished riskcv for permuted data.\n", file=stderr())
      cat("Finished riskcv for permuted data.\n")
    } else {
	   statprm <- pval.roc <- aucprm <- NULL
	}   # end of if (nperm > 0)

  # options(show.error.messages = TRUE, warn = 0)
  # cat("Complete risk group prediction.\n", file=stderr())
  cat("Complete risk group prediction.\n")
  permpval <- if (nperm > 0) pval else  NULL
  # statobs and statprm are not used in GetRiskResults.R. Debugging purpose; eg compare
  #         out$statobs$stat3 and out$statprm.
  # statobs, statprm are objects returned from riskcv().
  # labpred is an object returned from riskfull().
  modelE <- labpred$modelE
  modelE$cvperct <- statobs$predperct[,1]
  modelE$CVsupport <- statobs$CVsupport[,1]
  modelE$cvpred <- statobs$pred[,1]
  modelE$markervalue <- statobs$newprogindex[,1]
  modelCl <- labpred$modelCl
  # modelCl$cvperct <- statobs$predperct[,2] obsolute
  modelCl$cvpred <- statobs$pred[,2]
  modelCl$markervalue <- statobs$newprogindex[,2]
  modelCo <- labpred$modelCo
  modelCo$cvperct <- statobs$predperct[,3]
  modelCo$CVsupport <- statobs$CVsupport[,2]
  modelCo$cvpred <- statobs$pred[,3]
  modelCo$markervalue <- statobs$newprogindex[,3]
  return(list(err = err, permpval = permpval, IsImpute = IsImpute, Smallnlevel = (out.CC$nlevel <= 4),
              statprm = statprm, aucprm = aucprm, aucobs = aucobs, statobs = statobs, pval.roc = pval.roc,
              modelE = modelE, modelCl = modelCl, modelCo = modelCo, lambda = labpred$lambda, errmsg=errmsg))
}

#' Plot Kaplan-Meier curves for risk groups obtained from cross validation.
#'
#' This function plots Kaplan-Meier curves for risk groups obtained from cross validation
#' using models with gene expression only, with covariates only and with both gene expression
#' and covariates. If no covariates are provided, only the Kaplan-Meier curve related with the
#' model with gene expression only can be plotted. If any covariates are provided, Kaplan-Meier
#' curves for three models can be given.
#'
#' @param list list returned by function \code{survRiskPred}.
#' @param model numeric for a model considering
#' \itemize{
#'    \item 1:  gene expression only
#'    \item 2:  covariates only
#'    \item 3:  both gene expression and covariates
#' }
#' @export
#' @examples
#' res <- test.survRiskPredict("Pomeroy")
#' plotKMCurve(res, 1)
#' plotKMCurve(res, 2)
#' plotKMCurve(res, 3)

plotKMCurve <- function(list, model){
  if (model == 1){
    load(paste(list$workPath, "/plotKM.rda", sep=""))
    main = "Fitting by gene expression only"
  } else if (model == 2 ){
    load(paste(list$workPath, "/plotKMclin.rda", sep=""))
    main = "Fitting by covariates only"
  } else if (model == 3 ){
    load(paste(list$workPath, "/plotKMcomb.rda", sep=""))
    main = "Fitting by both gene expression and covariates"
  } else {
    stop("The values for fit can only be 1, 2 or 3. \n")
  }
  y <- y_plotKM
  tme <- tme_plotKM
  status <- status_plotKM
  riskgroup = TRUE
  cex.main = 1.0
  stat <- NA
  if (length(y) > 0) {
    ystring <- sort(unique(as.character(y)))
    y <- factor(y)
    uy <- unique(y)
    #if (length(unique(y)) >= 2) {
    stat <- if (length(ystring) > 1) survival::survdiff(survival::Surv(tme, status) ~ y)$chisq else 0
    #  if (!is.null(filename)) {
    #    png(filename)
    par(mar=c(5,4,4,2)+.1)
    if (length(uy) == 2 & riskgroup) {
      col <- c("red", "blue")
      legend.text <- c("High risk", "Low risk")
      lty <- 1:2
    } else if (length(uy) == 3 & riskgroup){
      col <- c("red", "green", "blue")
      # strata is sorted by alphabet.
      legend.text <- c("High risk", "Low risk", "Medium risk")
      lty <- 1:3
    } else if (length(uy) == 1 & riskgroup) {
      # added 9/26/06
      col <- ifelse (uy == "high", "red", "blue")
      legend.text <- ifelse(uy == "high", "High risk", "Low risk")
      lty <- 1
    } else if (!riskgroup) {
      legend.text <- ystring
      col <- rainbow(length(ystring))
      lty <- 1:length(ystring)
    }
    plot(survival::survfit(survival::Surv(tme, status) ~ y), conf.int = FALSE,
         lty = lty, col = col,
         xlab = "time", ylab = "Survival function")
    legend("bottomleft", legend=legend.text, col=col, lty=lty, bty="n")
    title(main = main, cex.main = cex.main)
    #    dev.off()
    #  }
    #}
  }
}


#' Plot ROC curves.
#'
#' This function plots the ROC curves at a specified landmark time using the Nearest Neighbor Estimation (NNE)
#' method of Heagerty, Lumley and Pepe, 2000. If no covariates are provided, only the ROC curve
#' with the model considering gene expression only will be plotted. If any covariates are provided,
#' two ROC curves with models considering gene expression only and considering both gene expression
#' and covariates will be given.
#'
#' @param list list returned by function \code{survRiskPred}.
#' @export
#' @examples
#' res <- test.survRiskPredict("Pomeroy")
#' plotROCCurveRisk(res)

plotROCCurveRisk <- function(list){
  load(paste(list$workPath, "/plotROCCurveModelE.rda", sep=""))
  aucobs <- rep(NA, 3)
  mdlroc <- seq(1,3)[sapply(1:3, function(i) all(is.finite(statobs$newprogindex[, i])))]
  plotted <- FALSE
  if (length(mdlroc) > 0) {
    for(i in mdlroc) {
      rocout <- try(survivalROC::survivalROC(Stime=tmetrain,  status=statustrain,  marker = statobs$newprogindex[, i],
                                             predict.time = landmarktime, span = 0.25*nsamp^(-0.20) ))
      if (!inherits(rocout, "try-error")) {
        if (plotted) par(new = TRUE)
        aucobs[i] <- rocout$AUC
        plot(rocout$FP, rocout$TP, type="l", xlim=c(0,1), ylim=c(0,1),  lty=i, xlab="", ylab="",
             main =paste("Landmark time:", landmarktime))
        plotted <- TRUE
      }
    }
    abline(0, 1, col="gray")
    aucval <- paste("AUC=", paste(round(aucobs[mdlroc], 3), c("(genes)", "(covariates)", "(combined)")[mdlroc], collapse=",", sep=""))
    title(xlab=paste( "1 - specificity", "\n", aucval), ylab="sensitivity")
    legend("topleft", lty=mdlroc, c("genes", "covariates", "combined")[mdlroc])

  }
}

# Convert the R script to a function for batr.
#
# Require <ArrayTools/Fortran/expernames> which contains a vector of experiment names of ALL samples.

GetRiskResults <- function(projectPath, analysisName, arraytoolsPath, WorkPath,
                           status, tme, training, arr1, nperm, nfold, clin, perc,
                           alpha, ncomp, uselasso, mixing, pcrgenes, landmarktime,
                           GenesFiltered, AnalysisType, generateHTML) {
  WorkPath <- file.path(projectPath, "Fortran", analysisName)

  UseFrames <- F
  IsMasterGene <- FALSE
  ExternalFileName <- outputdir <- paste(projectPath, "/Output/", analysisName, sep="")
  dir.create(dirname(ExternalFileName), showWarnings = FALSE)
  LoadingFileName <- paste(ExternalFileName,"/loading.html", sep = "")
  ExternalFileName <- paste(ExternalFileName, "/", analysisName, ".html", sep="")
  dir.create(dirname(ExternalFileName), showWarnings = FALSE)

  PlotKMFileName <- paste(projectPath, "/Output/", analysisName, "/", analysisName, " -PlotKM.png", sep="")
  unlink(PlotKMFileName)
  PlotrocFileName <- file.path(projectPath, "Output", analysisName, "rocModelE.png")
  unlink(PlotrocFileName)
  if (!is.null(clin)) {
    PlotKMFileNameclin <- paste(projectPath, "/Output/", analysisName, "/", analysisName, " -PlotKMclin.png", sep="")
    PlotKMFileNamecomb <- paste(projectPath, "/Output/", analysisName, "/", analysisName, " -PlotKMcomb.png", sep="")
    unlink(PlotKMFileNameclin)
    unlink(PlotKMFileNamecomb)
  } else {
    PlotKMFileNameclin <- PlotKMFileNamecomb <- NULL
  }

  PredictIndexFile <- paste(projectPath, "/Output/", analysisName, "/", analysisName, " -PlotPI.png", sep="")
  unlink(PredictIndexFile)
  if (!is.null(clin)) {
    PredictIndexFileclin <- paste(projectPath, "/Output/", analysisName, "/", analysisName, " -PlotPIcl.png", sep="")
    PredictIndexFilecomb <- paste(projectPath, "/Output/", analysisName, "/", analysisName, " -PlotPIco.png", sep="")
    unlink(PredictIndexFileclin)
    unlink(PredictIndexFilecomb)
  }
  # try(memory.limit(4000)) # for large data set. R 2.9.0 user should upgrade to R 2.9.0 patch
  system.time(out <- survRiskMain(status, tme, training, arr1, nperm, nfold, clin, perc,
                              alpha, ncomp, uselasso, mixing, pcrgenes, landmarktime,
                              PlotKMFileName, PlotKMFileNameclin, PlotKMFileNamecomb,
                              PredictIndexFile, PredictIndexFileclin, PredictIndexFilecomb,
                              PlotrocFileName, WorkPath)) #09/14

  if (out$err == 1) {
    if (generateHTML)
      cat("The number of levels for some categorical covariates is less than two. Please correct it and run the analysis again.",
                file = ExternalFileName) else
      stop("The number of levels for some categorical covariates is less than two. Please correct it and run the analysis again. \n")
  } else if (out$err == 2) {
    if (generateHTML)
      cat("The data contain missing values. The required R package can't be downloaded. Please check your internet connect.",
                file = ExternalFileName) else
      stop("The data contain missing values. The required R package can't be downloaded. Please check your internet connect. \n")
  } else if (out$err == 3) {
    if (generateHTML)
      cat(paste("KNN imputation stops because at least one column ",
                                     "of data has more than 80% missing values. ",
                                     "Make sure no arrays have more than 80% missing values ",
                                     "and rerun the analysis.", sep = ""),
                file = ExternalFileName) else
      stop(paste("KNN imputation stops because at least one column ",
           "of data has more than 80% missing values. ",
           "Make sure no arrays have more than 80% missing values ",
           "and rerun the analysis. \n", sep = "") )
  } else if (out$err == 4) {
    if (generateHTML)
      cat(paste("Missing data was still found after the imputation. It may be some genes have missing values on all arrays. ",
                                     "Please subset genes and rerun the analysis.", sep = ""),
                file = ExternalFileName) else
      stop(paste("Missing data was still found after the imputation. It may be some genes have missing values on all arrays. ",
                 "Please subset genes and rerun the analysis. \n", sep = ""));
  } else if (out$err == 5) {
    if (generateHTML)
      cat(paste("The model is not estimable with the covariates in the training set.",
                  "The analysis is aborted now.",
                  "Please double check the covariates in the training set and rerun the analysis again!"),
                file = ExternalFileName) else
      stop(paste("The model is not estimable with the covariates in the training set.",
                 "The analysis is aborted now.",
                 "Please double check the covariates in the training set and rerun the analysis again! \n"))
  } else if (out$err == 6) {
      if (!uselasso) {
        if (generateHTML)
          cat("No significant genes were found at the selected alpha level. Please increase the alpha value and ",
              "run the analysis again.", file = ExternalFileName) else
          stop("No significant genes were found at the selected alpha level. Please increase the alpha value and ",
               "run the analysis again. \n")
      } else  {
        if (generateHTML)
          cat("No genes were found based on the option in the penalized Cox regression. Please change the mixing parameter",
  		    "and run the analysis again.", file = ExternalFileName) else
  		      stop("No genes were found based on the option in the penalized Cox regression. Please change the mixing parameter",
  		           "and run the analysis again. \n")
      }
  } else if (out$err ==7) {
    if (generateHTML)
      cat("The glmnet function experiences a problem. The problem could be data specific. Please zip your whole project and send it
  	     arraytools@emmes.com for an investigation.", file = ExternalFileName) else
  	  stop("The glmnet function experiences a problem. The problem could be data specific. Please zip your whole project and send it
  	       arraytools@emmes.com for an investigation. \n")
  } else if (out$err == 0) {
      modelE <- out$modelE
      modelCl <- out$modelCl
      modelCo <- out$modelCo
      modelE$cvpred[modelE$cvpred=="" | is.na(modelE$cvpred)] <- "NA"
      modelCl$cvpred[modelCl$cvpred=="" | is.na(modelCl$cvpred)] <- "NA"
      modelCo$cvpred[modelCo$cvpred=="" | is.na(modelCo$cvpred)] <- "NA"
      RiskoutName <- paste(WorkPath, "/riskout.rda", sep="/") #Use WorkPath, Yong 12/15/06
      unlink(RiskoutName)
      save(out, file = RiskoutName)
      IsMasterGene <- F # Yong 8/21/06
      #Create genelist folder.
      dir.create(paste(projectPath, "/Genelists", sep=""), showWarnings=F)
      dir.create(paste(projectPath, "/Genelists/AnalysisResults", sep=""), showWarnings=F)
      unlink(ExternalFileName)
      load(paste(WorkPath, "/experiments.rda", sep="")) #use experiments.rda after modifying WriteFilesRiskPrediction.R, and use load/save method, Yong 12/6/06
      experimentsNew <- experiments[training == 0]
      experimentsTrain <- experiments[training == 1]

      ntest <- dim(arr1)[2] - sum(training)
      if (generateHTML){
      cat(  "<!-- This is an HTML document.  Please make sure the filename has either"
          , "     an .htm or .html extension, and view this file in a web browser. -->"
          , ""
          , "<HTML>"
          , "<HEAD>"
          , "<STYLE>"
          , "P {font-size: 9pt}"
          , "TABLE {font-size: 9pt}"
          , "</STYLE>"
          , paste("<script src='file:///", arraytoolsPath, "/Misc/sorttable.js'></script>", sep="") #Qian 5/29/09 sortable
          , "</HEAD>"
          , "<BODY>"
          , "<SCRIPT>"
          , "function popUp(text) {"
          , "   var prop ="
          , "   \"location=no,scrollbars=yes,menubars=no,toolbars=no,resizable=yes\";"
          , "   popup = window.open(\"\",\"mywin\",prop);"
          , "   popup.document.open();"
          , "   popup.document.write(text);"
          , "   popup.document.close();"
          , "   popup.focus();"
          , "   }"
          , "   </SCRIPT>"
          , "<TITLE>Risk Group Prediction Results</TITLE></HEAD>"
          , "<H2 align=\"center\"> Survival Risk Prediction</H2>"
          , "<H3>Contents </H3>"
          , "<ul>"
          , "<li><a href=\"#Description\"><b>Description of the problem</b></a></li>"
          , "<li><a href=\"#ModelE\"><b>Consider gene expression only</b></a></li>"
          , "  <ul>"
          , if (file.access(PlotKMFileName) == 0) "  <li><a href=\"#KaplanE\"><b>Kaplan-Meier curves for risk groups obtained from cross validation</b></a></li>"
          , if (ntest > 0 & length(modelE$ind.genes) > 0 && all(!is.na(modelE$coxcoef))) "  <li><a href=\"#PredE\"><b>Risk group predictions of new samples</b></a></li>"
          , if (length(modelE$ind.genes) > 0) "  <li><a href=\"#GenesE\"><b>Genes used in classifier of risk groups</b></a></li>"
          , if (length(modelE$ind.genes) > 0 & !uselasso) "  <li><a href=\"#LoadingE\"><b>Loading matrix of significant genes and
                             the correlations between the principal components and the signficant genes</b></a></li>"
          , if (any(modelE$cvpred != "NA")) "  <li><a href=\"#PerctE\"><b>Table of arrays, survival time, censoring indicator and survival risk prediction of training samples</b></a></li>"
          , "  </ul>"
          , if (!is.null(clin)) {
             paste("<li><a href=\"#ModelCl\"><b>Consider covariates only</b></a></li>"
             , "  <ul>"
             , if (file.access(PlotKMFileNameclin) == 0) {
                  if (!out$Smallnlevel) "  <li><a href=\"#KaplanCl\"><b>Kaplan-Meier curves for risk groups obtained from cross validation</b></a></li>"  else
                          "  <li><a href=\"#KaplanCl\"><b>Kaplan-Meier curves for groups defined by the covariates</b></a></li>"
                  }
             , if (ntest > 0 && !is.null(modelCl$pred) && !is.na(modelCl$pred)) "  <li><a href=\"#PredCl\"><b>Risk group predictions of new samples</b></a></li>"
             , if (any(modelCl$cvpred != "NA")) "  <li><a href=\"#PerctCl\"><b>Table of arrays,  survival time, censoring indicator and survival risk prediction of training samples</b></a></li>"
             , "  </ul>"
             , "<li><a href=\"#ModelCo\"><b>Consider both gene expression and covariates</b></a></li>"
             , "  <ul>"
             , if (file.access(PlotKMFileNamecomb) == 0) "  <li><a href=\"#KaplanCo\"><b>Kaplan-Meier curves for risk groups obtained from cross validation</b></a></li>"
             , if (ntest > 0 & length(modelCo$ind.genes) > 0 && all(!is.na(modelCo$coxcoef))) "<li><a href=\"#PredCo\"><b>Risk group predictions of new samples</b></a></li>"
             , if (length(modelCo$ind.genes) > 0) "  <li><a href=\"#GenesCo\"><b>Genes used in classifier of risk groups</b></a></li>"
             , if (length(modelCo$ind.genes) > 0 & !uselasso) "  <li><a href=\"#LoadingCo\"><b>Loading matrix of significant genes and
                  the correlations between the principal components and the signficant genes</b></a></li>"
             , if (any(modelCo$cvpred != "NA")) "  <li><a href=\"#PerctCo\"><b>Table of arrays,  survival time, censoring indicator and survival risk prediction of training samples</b></a></li>"
             , "  </ul>"
             , sep = "\n")
            }
  		    , if (file.exists(PlotrocFileName)) " <li><a href=\"#rocE\"><b>Time-dependent ROC curve</b></a></li>"
  		    , "</ul>"
          , "<P></P>", "<HR SIZE=5 WIDTH=\"100%\" NOSHADE>"
          , paste("<H3><a name=\"Description\"><B><U>Description of the problem:</U></B></a></H3>")
          , "<P>"
          , paste("Number of genes that passed filtering criteria:",GenesFiltered )
          , if (!uselasso) {
              paste("<BR>", paste("Number of principal component(s):", ncomp), "<BR>", paste("Number of risk groups:", length(perc)+1))
            } else {
  		        paste("<BR>\nElastic-net mixing parameter:", mixing)
  		      }
          , "<BR>"
          , paste("Percentile of prognostic index in constructing risk groups: ", paste(perc, collapse = ", "), " <BR>", sep = "")
          , if (nperm > 0) paste("Number of permutations for the log-rank test statistic:", nperm, "<BR>")
          , paste("Cross-validation method:", ifelse(nfold ==10, "10-fold CV", "Leave-one-out"), "<BR>")
          , if (out$IsImpute) paste("10-Nearest Neighbor method is used to impute missing values in log expression", "<BR>")
          # , if (CovariateVariable != "") paste("Columns of the Experiment Descriptors sheet that define covariate variables:",paste(strsplit(CovariateVariable,',')[[1]], collapse=", "), "<BR>")
          #, if (sum(training == 0) > 0) paste("Column of the Experiment Descriptors sheet that defines training/prediction variable:",TrainingVariable, "<BR>")
  		    , if (uselasso) paste("Desired number of genes:", pcrgenes, "<BR>")
  		    , if (file.exists(PlotrocFileName)) paste("Landmark time for time-dependent ROC curve:", landmarktime, "<BR>\n") else "\n"
          ," ", file=ExternalFileName, sep="\n", append=F)
      cat("<P></P>", "<HR SIZE=5 WIDTH=\"100%\" NOSHADE>", "<H3>",
          "<a name=\"ModelE\"><B>Consider gene expression only</B></H3>", file=ExternalFileName, sep="\n", append=T)
      if (length(modelE$ind.genes) == 0) {
  	    if (!uselasso) {
        			cat("<P>No significant genes were found based on alpha equal to ", alpha, "</P>\n",
  			      	file=ExternalFileName, append=TRUE)
  	    } else {
  	    	  	cat("<P>No significant genes were found based on penalized Cox proportional hazards model </P>\n",
  			  	  file=ExternalFileName, append=TRUE)
  	    }
      }

      if (file.access(PlotKMFileName)==0 && length(modelE$ind.genes) > 0) { ##Ting 04/22/15 #If file is read-accessible.
          cat("<P><a name=\"KaplanE\"><B><U>Kaplan-Meier curves for risk groups obtained from cross validation:</U></B></a><BR>"
              , file=ExternalFileName, sep="\n", append=T)
          if (nperm > 0 && !is.na(out$permpval[1])) cat(paste("Permutation P-value of the log-rank test statistic between risk groups based on ",
                                     nperm, " permutations: ", format.pval(round(out$permpval[1],5), eps = 1/nperm), "<BR>")
                                    , file=ExternalFileName, sep="\n", append=T)
          cat(paste("<IMG SRC=\"", analysisName, " -PlotKM.png", "\">", "</P>", sep="")
            ,file=ExternalFileName, sep="\n", append=T)
      }
      } #end of if (generateHTML)

  	  if (length(modelE$ind.genes) == 0 && file.exists(PlotKMFileName))
  		  file.remove(PlotKMFileName)

      # Predictions for new samples:
      if(ntest > 0 & length(modelE$ind.genes) > 0 && all(!is.na(modelE$coxcoef))) {
         prediction.table <- data.frame(ExpID = experimentsNew,
                                # stime = tme[training==0], status = status[training==0],
                                newprogindex = round(modelE$newprogindex, 3), Prediction = as.character(modelE$pred))

         # for returning a list
         prediction.table2 <- prediction.table

         if (generateHTML){
         prediction.table$ExpID <- paste("<TR ALIGN=center><TH>", prediction.table$ExpID )
         ExpClassifiedNew <- rep( "", ntest+1 )
         ExpClassifiedNew[1] <- paste("<BR><TH>Array id <TH> Prognostic index <TH WIDTH=65>Prediction")
         prediction.table <- as.matrix(prediction.table)
         for(i in 1:ntest) {
             ExpClassifiedNew[i+1] <- paste(prediction.table[i,,drop=F], collapse="<TD>")
         }
         cat("<HR SIZE=1 WIDTH=\"100%\" NOSHADE>"
          , "<P><a name=\"PredE\"><B><U>Risk group predictions of new samples:</U></B></a></P>"
          , "<table border='1' class='sortable'>" #Qian 5/29/09 sortable
          , ExpClassifiedNew
          , "</TABLE>"
          , " "
          , file=ExternalFileName, sep="\n", append=T)
         } #end of if (generateHTML)

         # for returning a list
         names(prediction.table2) <- c("Array id", "Prognostic index", "Prediction")
         resList <- list(predictNewSamplesModE=prediction.table2)
      }
      # End of if(ntest > 0 & !is.na(modelE$pred)) part

	  ppval <- NULL
	  namesppval <- NULL
	  if (nperm > 0 && !is.null(out$permpval[1])){
		ppval <- c(ppval, format.pval(round(out$permpval[1],5), eps = 1/nperm))
		namesppval <- c(namesppval, "ModelE")
	  }
	  if (nperm > 0 && !is.null(out$permpval[2])){
		ppval <- c(ppval, format.pval(round(out$permpval[2],5), eps = 1/nperm))
		namesppval <- c(namesppval, "ModelCl")
	  }	  
	  if (nperm > 0 && !is.null(out$permpval[3])){
		ppval <- c(ppval, format.pval(round(out$permpval[3],5), eps = 1/nperm))
		namesppval <- c(namesppval, "ModelCo")
	  }	  
	  names(ppval) <- namesppval
	  resList <- c(resList, list(ppval=ppval))
	  
      # WRITE gene table:
      ncomp2 <- min(ncomp, length(modelE$ind.genes))
      CVsupport <- modelE$CVsupport[modelE$ind.genes]

      if (length(modelE$ind.genes) > 0) {
  	    if (!uselasso) {
  		    if (is.null(clin)) { # set DoMasterAnnotations=True, Yong 8/21/06
  			    GetGenesHTMLOutput <- GetGenesHTML(index=modelE$ind.genes, puniv=modelE$pval, DoMasterAnnotations=T,
  				 								genelistfilename = GeneListFileName1, WorkPath=WorkPath, AnalysisType = AnalysisType,
  				 								uselasso = uselasso, CVsupport = CVsupport)
  		    } else { # set DoMasterAnnotations=False, Yong 8/21/06
  			    IsMasterGene <- T
  			    Master.ind <- modelE$ind.genes
  			    GetGenesHTMLOutput <- GetGenesHTML(index=modelE$ind.genes, puniv=modelE$pval, DoMasterAnnotations=F,
  												genelistfilename = GeneListFileName1, WorkPath=WorkPath, AnalysisType = AnalysisType,
  												uselasso = uselasso, CVsupport = CVsupport)
  		    }
  	    } else {
  	      CVsupport <- modelE$CVsupport[modelE$ind.genes]
  		    coxcoef <- modelE$coxcoef # used in GetGenesHTML
  		    if (is.null(clin)) {
  			    GetGenesHTMLOutput <- GetGenesHTML(index=modelE$ind.genes, DoMasterAnnotations=T,
  												genelistfilename = GeneListFileName1, WorkPath=WorkPath, AnalysisType = AnalysisType,
  												uselasso = uselasso, CVsupport = CVsupport, coxcoef = coxcoef) # Ting added coxcoef 9/8/2017
  		    } else {
  			    IsMasterGene <- T
  			    Master.ind <- modelE$ind.genes
  			    GetGenesHTMLOutput <- GetGenesHTML(index=modelE$ind.genes, DoMasterAnnotations=F,
  												genelistfilename = GeneListFileName1, WorkPath=WorkPath, AnalysisType = AnalysisType,
  												uselasso = uselasso, CVsupport = CVsupport, coxcoef = coxcoef) # Ting added coxcoef 9/8/2017
  		    }
  	    } # end of if (!uselasso)

        if (generateHTML){
        cat("<P></P>"
          , "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>"
          , "<P><B><a name=\"GenesE\"><U>Genes used in classifier of risk groups: </U></B></a><BR>"
          , file=ExternalFileName, sep="\n", append=T)
  	    if (!uselasso) {
  			  cat(length(modelE$ind.genes), " genes selected by fitting Cox proportional hazards models (alpha equals to "
    				, signif(alpha,3),")<BR>"
    				, paste("The coefficients of the fitted Cox proportional hazards model using the principal components
    						from the training dataset is (", paste(round(modelE$coxcoef, 3), collapse=", ") ,
    						")<BR>", sep = "")
    				, file=ExternalFileName, sep="\n", append=T)
    			cat(paste("The percent of variability explained by the first ", ncomp2, "principal",
  					if (ncomp2 >1) "components" else "component", "is",
  					round(modelE$percent, 3), "<BR>")
    				, file=ExternalFileName, sep="\n", append=T)
  	  		cat(paste("The p-value in the table is testing the hypothesis if expression data is preditive of survival.<BR></P>")
  		  		, paste("<P><B>Table</B></P>")
    				, GetGenesHTMLOutput$GeneId1 #Qian 6/3/09
    				, file=ExternalFileName, sep="\n", append=T)
  	    } else {
  		    cat(length(modelE$ind.genes), " genes selected by fitting penalized Cox proportional hazards model <BR>",
  			      file=ExternalFileName, sep="\n", append=T)
  			  cat(paste("<P><B>Table</B></P>")
    				, GetGenesHTMLOutput$GeneId1
    				, file=ExternalFileName, sep="\n", append=T)
  	    } # end of if (!uselasso)
        } #end of if (generateHTML)

        # for returning a list
        resList <- c(resList, list(genesInClassifierModE=GetGenesHTMLOutput$GeneIdForIngenuity))
      } # end of if (length(modelE$ind.genes) > 0)

      # Write Loading matrix
      if (length(modelE$ind.genes) > 0 & !uselasso) {
        if (generateHTML)
        cat("<P></P>", "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>",
            "<P></P><A name=\"LoadingE\"><B><U>Loading matrix of the significant genes and
            the correlations between the principal components and the signficant genes.</U></B></A>",
            if (length(perc)==1) paste("<P>A new sample is predicted as high (low) risk if its prognostic index is larger than (smaller than or equal to) ", round(modelE$threshold,6), ".</P>", sep="") else paste("<P>A new sample is predicted as high risk if its prognostic index is larger than ", round(modelE$threshold[2],6), ", medium risk if the index is within (", round(modelE$threshold[1],6), ", ", round(modelE$threshold[2],6), "], low risk if the index is smaller than or equal to ", round(modelE$threshold[1],6), ".</P>", sep=""),
            "<P>The prognostic index can be computed by the simple formula",
            "&#8721;<sub><var>i</var></sub><var>w<sub>i</sub></var>
            <var>x<sub>i</sub></var> ", ifelse(modelE$indexconst >0, '- ', '+ '),
            abs(round(modelE$indexconst, 6)),
            " where <var>w<sub>i</sub></var> and <var>x<sub>i</sub></var> are the weight and logged gene expression for the <var>i</var>-th gene.</P>",
            file=ExternalFileName, sep="\n", append=T)
        # Obtain GeneId1
        # GeneIdNames <- dget(paste(arraytoolsPath, "/Fortran/GeneIdNames.rda", sep=""))
  #      NumGeneIds <- length(dget(paste(arraytoolsPath, "/Fortran/GeneIdNames.rda", sep="")))
  #      GeneId1 <- dget(paste(arraytoolsPath, "/Fortran/GeneId1.rda", sep=""))
        load(paste(WorkPath, "/GeneIdNames.rda", sep="")) #use load/save method, Yong 12/6/06. Use WorkPath 12/15/06
        NumGeneIds <- length(GeneIdNames)
        load(paste(WorkPath, "/GeneId1.rda", sep=""))
        #Qian 6/3/09
        #ExpressionTable.H1 <- paste("<TR><TH ROWSPAN=3>#<TH ROWSPAN=3>Gene Id<TH COLSPAN=",
        #                        ncomp2,">Loading matrix<TH COLSPAN=",ncomp2,">Correlation<TH ROWSPAN=3>Weights (<var>w<sub>i</sub></var>)</TR>")
        #ExpressionTable.H2 <- paste("<TR><TH COLSPAN=",ncomp2,">Components<TH COLSPAN=",ncomp2,">Components</TR>")
        #ExpressionTable.H3 <- paste("<TR><TH>", paste(1:ncomp2, collapse="<TH>"),"<TH>", paste(1:ncomp2, collapse="<TH>"), "</TR>")
        ExpressionTable.H1 <- "<TR><TH>&nbsp;<TH>Gene Id"
        ExpressionTable.H2 <- paste("<TH>Loading", paste(1:ncomp2, collapse="<TH>Loading"),"<TH>Correlation", paste(1:ncomp2, collapse="<TH>Correlation"))
        ExpressionTable.H3 <- "<TH>Weights (<var>w<sub>i</sub></var>)</TR>"
        #TableTitle1 <- "The loading matrix of significant genes and the correlations
        #            between the significant genes and the principal components for
        #            the Survival Risk Prediction Analysis."
        nHeaderLines  <- 3
        ExpressionData <- cbind(modelE$loading, modelE$gpcor, modelE$loading %*% modelE$coxcoef)
        order.pvalue <- order(modelE$pval)
        #cat(order.pvalue, file = file.path(WorkPath, "ordergenesmodelE.txt"))
        if (generateHTML){
        ExpressionTable <- character(nHeaderLines+length(ExpressionData[,1]) )
        ExpressionTable[1] <- ExpressionTable.H1
        ExpressionTable[2] <- ExpressionTable.H2
        ExpressionTable[3] <- ExpressionTable.H3
        #EndQian
        for( i in 1:length(ExpressionData[,1])) {
           ExpressionTable[i+nHeaderLines] <- paste("<TR><TH>",i,"<TH>",
               GeneId1[modelE$ind.genes[order.pvalue[i]]],"<TD NOWRAP ALIGN=\"RIGHT\">",
               paste(round(ExpressionData[order.pvalue[i],],6) ,collapse="<TD NOWRAP ALIGN=\"RIGHT\">"))
        }
        # the <P></P> below can create a space.
        cat("<P></P>", "<table border='1' class='sortable'>" #Qian 5/29/09 sortable
        # , paste("<P>Table.",TableTitle1,"</P>")
         , "<CAPTION ALIGN=left></CAPTION>"
         , ExpressionTable
         , "</TABLE>"
         , file=ExternalFileName, sep="\n", append= TRUE)
        } #end of if (generateHTML)

        # for returning a list
        ExpressionTable2 <- matrix(0, dim(ExpressionData)[1],dim(ExpressionData)[2] )
        for( i in 1:length(ExpressionData[,1]))
          ExpressionTable2[i,] <- round(ExpressionData[order.pvalue[i],],6)
        rownames(ExpressionTable2) <- GeneId1[modelE$ind.genes[order.pvalue]]
        names <- NULL
        for (i in 1:ncomp2)
          names <- c(names, paste("Loading ",i,sep=""))
        for (i in 1:ncomp2)
          names <- c(names, paste("Correlation ",i,sep=""))
        colnames(ExpressionTable2) <- c(names, "Weights")
        resList <- c(resList, list(loadingMatrixModE=ExpressionTable2))

      }
      # Write table of arrays, survival time, censoring indicator and the predicted percentile
      if (any(modelE$cvpred != "NA" && length(modelE$ind.genes) > 0)) { ## Ting 04/22/15
        if (generateHTML)
        cat("<P></P>", "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>",
            "<P></P><A name=\"PerctE\"><B><U>Table of arrays, survival time, censoring indicator and survival risk prediction of training samples.</U></B></A>",
            file=ExternalFileName, sep="\n", append=T)
        ExpressionData <- data.frame(array = as.character(experimentsTrain),
                                     stime = round(tme[training == 1],3),
                                     status = status[training == 1], # perct = round(modelE$cvperct,1),
                                     pred = as.character(modelE$cvpred))
         colnames(ExpressionData) <- c("Array", "Survival time", "Censoring indicator",
                                        #"Predicted percentile",
                                        "Predicted risk")

         # for returning a list
         resList <- c(resList, list(predictRiskTrainingModE=ExpressionData))
         if (generateHTML){
         ExpressionTable <- HTML.data.frame.my(ExpressionData)
         cat("<P></P>", ExpressionTable, "\n", file=ExternalFileName, append= TRUE)
         } #end of if (generateHTML)
      }
      if (!is.null(clin)) {
      #  ModelCl
        if (generateHTML){
         cat("<P></P>", "<HR SIZE=5 WIDTH=\"100%\" NOSHADE>",
             "<H3><a name=\"ModelCl\"><B>Consider covariates only</B></H3>", file=ExternalFileName, sep="\n", append=T)
         if(file.access(PlotKMFileNameclin)==0)  {
             if (!out$Smallnlevel) {
               cat("<P><a name=\"KaplanCl\"><B><U>Kaplan-Meier curves for risk groups obtained from cross validation:</U></B></a><BR>"
                  , file=ExternalFileName, sep="\n", append=T)
             } else {
               cat("<P><a name=\"KaplanCl\"><B><U>Kaplan-Meier curves for groups defined by the covariates:</U></B><BR>
                     Each curve corresponds to one level in the covariates</a><BR>"
                  , file=ExternalFileName, sep="\n", append=T)
             }
             if (nperm > 0 && !is.na(out$permpval[2])) {
                cat(paste("P-value of the parametric log-rank test statistic between risk groups: ",
                          round(out$permpval[2],5), "<BR>")
                          , file=ExternalFileName, sep="\n", append=T)
             }
             cat(paste("<IMG SRC=\"", analysisName, " -PlotKMclin.png", "\">", "</P>", sep="")
               ,file=ExternalFileName, sep="\n", append=T)
         }
        } #end of if (generateHTML)
         # Predictions for new samples:
         if(ntest > 0 && !is.null(modelCl$pred) && !is.na(modelCl$pred)) {
            # repeat the above code with modelCl.
            prediction.table <- data.frame(ExpID = experimentsNew,
                                #stime = tme[training==0], status = status[training==0],
                                newprogindex = round(modelCl$newprogindex, 3), Prediction = as.character(modelCl$pred))

            # for returning a list
            prediction.table2 <- prediction.table

            if (generateHTML){
            prediction.table$ExpID <- paste("<TR ALIGN=center><TH>", prediction.table$ExpID )
            ExpClassifiedNew <- rep( "", ntest+1 )
            ExpClassifiedNew[1] <- paste("<BR><TH>Array id <TH> Prognostic index <TH WIDTH=65>Prediction")
            prediction.table <- as.matrix(prediction.table)
            for(i in 1:ntest) {
                ExpClassifiedNew[i+1] <- paste(prediction.table[i,,drop=F], collapse="<TD>")
            }
            cat("<HR SIZE=1 WIDTH=\"100%\" NOSHADE>"
             , "<P><B><a name=\"PredCl\"><U>Risk group predictions of new samples:</U></B></A></P>"
             , "<table border='1' class='sortable'>" #Qian 5/29/09 sortable
             , ExpClassifiedNew
             , "</TABLE>"
             , " "
             , file=ExternalFileName, sep="\n", append=T)
            } #end of if (generateHTML)

            # for returning a list
            names(prediction.table2) <- c("Array id", "Prognostic index", "Prediction")
            resList <- c(resList, list(predictNewSamplesModCl=prediction.table2))
         } # End of if(ntest) part
         # Write table of arrays, survival time, censoring indicator and the predicted percentile
         if (any(modelCl$cvpred != "NA" )) {
           if (generateHTML)
            cat("<P></P>", "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>",
                "<P></P><A name=\"PerctCl\"><B><U>Table of arrays, survival time, censoring indicator and survival risk prediction of training samples.</U></B></A>",
                file=ExternalFileName, sep="\n", append=T)
            ExpressionData <- data.frame(array = as.character(experimentsTrain),
                                         stime = round(tme[training == 1],3),
                                         status = status[training == 1],
                                         #perct = round(modelCl$cvperct, 1),
                                         pred = as.character(modelCl$cvpred))
            ExpressionData <- format(ExpressionData)
            ExpressionData <- as.matrix(ExpressionData)
            colnames(ExpressionData) <- c("Array", "Survival time", "Censoring indicator",
                                        #"Predicted percentile",
                                        "Predicted risk")

            # for returning a list
            resList <- c(resList, list(predictRiskTrainingModCl=ExpressionData))
            if(generateHTML){
            ExpressionTable <- HTML.data.frame.my(ExpressionData)
            cat("<P></P>", ExpressionTable, "\n", file=ExternalFileName, append= TRUE)
            } #end of if (generateHTML)
         }
         #  ModelCo
        if (generateHTML)
         cat("<P></P>", "<HR SIZE=5 WIDTH=\"100%\" NOSHADE>",
             "<H3><a name=\"ModelCo\"><B>Consider both gene expression and covariates</B></H3>", file=ExternalFileName, sep="\n", append=T)
         if (uselasso & length(out$lambda) < 2) {
           if (generateHTML)
  	      cat("<P>An error occurred in fitting the penalized Cox regression containing both gene expression and clinical covariates.</P>\n",
  		      file=ExternalFileName, append=TRUE)
          ErrLassoCo <- TRUE
  	   } else {
  	      ErrLassoCo <- FALSE
  	   }
       if (generateHTML){
       if (!ErrLassoCo && length(modelCo$ind.genes) == 0) {
  			  if (!uselasso) {
  				  cat("<P>No significant genes were found based on alpha equal to ", alpha, "</P>\n",
                      file=ExternalFileName, append=TRUE)
  			  } else {
  				  cat("<P>No significant genes were found based on penalized Cox proportional hazards model </P>\n",
  					    file=ExternalFileName, append=TRUE)
  			  }
  		  }
        if(file.access(PlotKMFileNamecomb)==0 && length(modelCo$ind.genes) > 0)  { ## Ting 04/22/15
             cat("<P><a name=\"KaplanCo\"><B><U>Kaplan-Meier curves for risk groups obtained from cross validation:</U></B></a><BR>"
              , file=ExternalFileName, sep="\n", append=T)
             if (nperm > 0  && !is.null(out$permpval[3])) {
                cat(paste("Permutation P-value of the combined model versus the covariates only model based on ",
                              nperm, " permutations: ", format.pval(round(out$permpval[3],5), eps = 1/nperm), "<BR>")
                              , file=ExternalFileName, sep="\n", append=T)
             }
             cat(paste("<IMG SRC=\"", analysisName, " -PlotKMcomb.png", "\">", "</P>", sep="")
               ,file=ExternalFileName, sep="\n", append=T)
        }
       } # end of if (generateHTML)
  	   if (length(modelCo$ind.genes) == 0 && file.exists(PlotKMFileNamecomb))
  		   file.remove(PlotKMFileNamecomb)
         # Predictions for new samples:
         if(ntest > 0 & length(modelCo$ind.genes) > 0 && all(!is.na(modelCo$coxcoef))) {
            # repeat the above code with modelCo.
            prediction.table <- data.frame(ExpID = experimentsNew,
                                #stime = tme[training==0], status = status[training==0],
                                newprogindex = round(modelCo$newprogindex, 3), Prediction = as.character(modelCo$pred))

            # for returning a list
            prediction.table2 <- prediction.table

            if (generateHTML){
            prediction.table$ExpID <- paste("<TR ALIGN=center><TH>", prediction.table$ExpID )
            ExpClassifiedNew <- rep( "", ntest+1 )
            ExpClassifiedNew[1] <- paste("<BR><TH>Array id <TH> Prognostic index <TH WIDTH=65>Prediction")
            prediction.table <- as.matrix(prediction.table)
            for(i in 1:ntest) {
                ExpClassifiedNew[i+1] <- paste(prediction.table[i,,drop=F], collapse="<TD>")
            }
            cat("<HR SIZE=1 WIDTH=\"100%\" NOSHADE>"
             , "<P><B><a name=\"PredCo\"><U>Risk group predictions of new samples:</U></B></A></P>"
             , "<table border='1' class='sortable'>" #Qian 5/29/09 sortable
             , ExpClassifiedNew
             , "</TABLE>"
             , " "
             , file=ExternalFileName, sep="\n", append=T)
            } # end of if (generateHTML)
            # for returning a list
            names(prediction.table2) <- c("Array id", "Prognostic index", "Prediction")
            resList <- c(resList, list(predictNewSamplesModCo=prediction.table2))
         }
         # Write significant gene table
         ncomp2 <- min(ncomp, length(modelCo$ind.genes))
         CVsupport <- modelCo$CVsupport[modelCo$ind.genes]
         if (length(modelCo$ind.genes) > 0) {
            #Master.ind <- union(Master.ind, modelCo$ind.genes) # Yong added 8/21/06
  		      if (!uselasso) {
  			      if (class(try(Master.ind, silent=T))=="try-error") { # Yong modified 11/7/06 to handle when there is no significant pathways/genes in gene model.
  				      Master.ind <- modelCo$ind.genes
  			      } else {
  				      Master.ind <- union(Master.ind, modelCo$ind.genes)
  			      }
  			      GetGenesHTMLOutput <- GetGenesHTML(index=modelCo$ind.genes, puniv=modelCo$pval, DoMasterAnnotations=F,
  												  genelistfilename = GeneListFileName2, WorkPath=WorkPath, AnalysisType = AnalysisType,
  												  uselasso = uselasso, CVsupport = CVsupport)
      	    } else {
  				    if (class(try(Master.ind, silent=T))=="try-error")  {
  					    Master.ind <- modelCo$ind.genes
  				    } else {
  				      Master.ind <- union(Master.ind, modelCo$ind.genes)
  				    }
  				    coxcoef <- modelCo$coxcoef # used in GetGenesHTML
  				    GetGenesHTMLOutput <- GetGenesHTML(index=modelCo$ind.genes, DoMasterAnnotations=F,
  														genelistfilename = GeneListFileName2, WorkPath=WorkPath, AnalysisType = AnalysisType,
  														uselasso = uselasso, CVsupport = CVsupport, coxcoef = coxcoef) # Ting added coxcoef 9/8/2017
      	    }
           if (generateHTML){
            cat("<P></P>"
              , "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>"
              , paste("<P><a name=\"GenesCo\"><B><U>Genes used in classifier of risk groups: </U></B></a><BR>", sep="\n")
  			      , file=ExternalFileName, sep="\n", append=T)
            if (!uselasso) {
      		    if (all(!is.na(modelCo$coxcoef))) {
  				      cat(length(modelCo$ind.genes), " genes selected by fitting Cox proportional hazards models (alpha equals to ",
  					        signif(alpha,3),")<BR>", file=ExternalFileName, sep="\n", append=T)
                cat(paste("The Cox proportional hazards model is fitted using the principal components and clinical covariates from the training dataset. The estimated coefficients are (",
                        paste(round(modelCo$coxcoef[1:ncomp2], 3), collapse=", ") ,
                        ") for the principal components and (",
                        paste(round(modelCo$coxcoef[-c(1:ncomp2)], 3), collapse=", ") ,
                        ") for the clinical covariates",
                        "<BR>", sep = ""), file=ExternalFileName, sep="\n", append=T)
              } else {
                    cat("Cox proportional hazards model estimation using the principal components of the full traing arrays and clinical covariates can not be completed <BR>\n",
  				      file=ExternalFileName, sep="\n", append=T)
              }
  				    cat(paste("The percent of variability explained by the first ", ncomp2, "principal",
  					      if (ncomp2 >1) "components" else "component", "is",
  					      round(modelCo$percent, 3), "<BR>")
  					      , file=ExternalFileName, sep="\n", append=T)
  			      cat(paste("The p-value in the table is testing the hypothesis if the expression data
  						    is predictive of survival over and above the covariates.<BR></P>")
  				      	, paste("<P><B>Table</B></P>")
  	      				, GetGenesHTMLOutput$GeneId1
  				      	, file=ExternalFileName, sep="\n", append=T)
  		      } else {
  				    cat(length(modelCo$ind.genes), " genes selected by fitting penalized Cox proportional hazards model <BR>",
  			          file=ExternalFileName, sep="\n", append=T)
      				cat(paste("<P><B>Table</B></P>")
  		    			, GetGenesHTMLOutput$GeneId1
  				    	, file=ExternalFileName, sep="\n", append=T)
  		      }
           } # end of if (generateHTML)

            # for returning a list
            resList <- c(resList, list(genesInClassifierModCo=GetGenesHTMLOutput$GeneIdForIngenuity))

         } # end of if (length(modelCo$ind.genes) > 0)

         # Write Loading matrix
         # If all covariates are continuous, the prediction rule will be printed out.
         if (length(modelCo$ind.genes) > 0 & !uselasso) {
            IsContinuousCov <- all(apply(clin, 2, is.numeric))
            if (generateHTML){
            tmpstr <- c("<P></P>", "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>",
               "<P></P><A name=\"LoadingCo\"><B><U>Loading matrix of the significant genes and
               the correlations between the principal components and the signficant genes.</U></B></A>")
            if (IsContinuousCov) {
                CovariateVariable2 <- strsplit(CovariateVariable, ',')[[1]]
                if (length(perc)==1) tmpstr <- c(tmpstr, "<P>A new sample is predicted as high (low) risk if its prognostic index is larger than (smaller than or equal to) ", round(modelCo$threshold,6), ".</P>") else tmpstr <- c(tmpstr, paste("<P>A new sample is predicted as high risk if its prognostic index is larger than ", round(modelCo$threshold[2],6), ", medium risk if the index is within (", round(modelCo$threshold[1],6), ", ", round(modelCo$threshold[2],6), "], low risk if the index is smaller than or equal to ", round(modelCo$threshold[1],6), ".</P>", sep=""))
                tmpstr <- c(tmpstr,
                            "<P>The prognostic index can be computed by the simple formula",
                            "&#8721;<sub><var>i</var></sub><var>w<sub>i</sub></var>
                             <var>x<sub>i</sub></var> ",
                            ifelse(modelCo$coxcoef[ncomp+1] >0, '+ ', '- '), abs(round(modelCo$coxcoef[ncomp+1], 6)),"<var>Cov<sub>1</sub></var>",
                            if(ncol(clin)>1) paste(ifelse(modelCo$coxcoef[ncomp+2] >0, '+ ', '- '), abs(round(modelCo$coxcoef[ncomp+2], 6)),"<var>Cov<sub>2</sub></var>"),
                            if(ncol(clin)>2) paste(ifelse(modelCo$coxcoef[ncomp+3] >0, '+ ', '- '), abs(round(modelCo$coxcoef[ncomp+3], 6)),"<var>Cov<sub>3</sub></var>"),
                            if(ncol(clin)>3) paste(ifelse(modelCo$coxcoef[ncomp+4] >0, '+ ', '- '), abs(round(modelCo$coxcoef[ncomp+4], 6)),"<var>Cov<sub>4</sub></var>"),
                            if(ncol(clin)>4) paste(ifelse(modelCo$coxcoef[ncomp+5] >0, '+ ', '- '), abs(round(modelCo$coxcoef[ncomp+5], 6)),"<var>Cov<sub>5</sub></var>"),
  							            if(ncol(clin)>5) paste(ifelse(modelCo$coxcoef[ncomp+6] >0, '+ ', '- '), abs(round(modelCo$coxcoef[ncomp+6], 6)),"<var>Cov<sub>6</sub></var>"),
                              ifelse(modelCo$indexconst >0, '- ', '+ '), abs(round(modelCo$indexconst, 6)),
                              " where <var>w<sub>i</sub></var> and <var>x<sub>i</sub></var> are the weight and logged gene expression for the <var>i</var>-th gene and <var>Cov<sub>j</sub></var> is the <var>j</var>-th covariate.")
                tmpstr <- c(tmpstr,
                           paste("<var>Cov<sub>1</sub></var>: ", CovariateVariable2[1],
                           if(ncol(clin)>1) paste(", <var>Cov<sub>2</sub></var>: ", CovariateVariable2[2]),
                           if(ncol(clin)>2) paste(", <var>Cov<sub>3</sub></var>: ", CovariateVariable2[3]),
                           if(ncol(clin)>3) paste(", <var>Cov<sub>4</sub></var>: ", CovariateVariable2[4]),
                           if(ncol(clin)>4) paste(", <var>Cov<sub>5</sub></var>: ", CovariateVariable2[5]),
  						             if(ncol(clin)>5) paste(", <var>Cov<sub>6</sub></var>: ", CovariateVariable2[6]),
                           ".", sep=""),
                              "</P>")
            }
            cat(tmpstr, file=ExternalFileName, sep="\n", append=T)
            ExpressionTable.H1 <- "<TR><TH>&nbsp;<TH>Gene Id"
            ExpressionTable.H2 <- paste("<TH>Loading", paste(1:ncomp2, collapse="<TH>Loading"),"<TH>Correlation", paste(1:ncomp2, collapse="<TH>Correlation"))
            if (IsContinuousCov) {
                ExpressionTable.H3 <- "<TH>Weights (<var>w<sub>i</sub></var>)</TR>"
            } else {
                ExpressionTable.H3 <- "</TR>"
            }
            } # end of if (generateHTML)
            nHeaderLines  <- 3
            ExpressionData <- cbind(modelCo$loading, modelCo$gpcor)
            if (IsContinuousCov) ExpressionData <- cbind(ExpressionData, modelCo$loading %*% modelCo$coxcoef[1:ncomp])
            order.pvalue <- order(modelCo$pval)
            if (generateHTML){
            ExpressionTable <- character(nHeaderLines+length(ExpressionData[,1]) )
            ExpressionTable[1] <- ExpressionTable.H1
            ExpressionTable[2] <- ExpressionTable.H2
            ExpressionTable[3] <- ExpressionTable.H3
            for( i in 1:length(ExpressionData[,1])) {
                ExpressionTable[i+nHeaderLines] <- paste("<TR><TH>",i,"<TH>",
                    GeneId1[modelCo$ind.genes[order.pvalue[i]]],"<TD NOWRAP ALIGN=\"RIGHT\">",
                    paste(round(ExpressionData[order.pvalue[i],],6) ,collapse="<TD NOWRAP ALIGN=\"RIGHT\">"))
            }
            cat("<P></P>", "<table border='1' class='sortable'>" #Qian 5/29/09 sortable
              , "<CAPTION ALIGN=left></CAPTION>"
              , ExpressionTable
              , "</TABLE>"
              , file=ExternalFileName, sep="\n", append = TRUE)
            } # end of if (generateHTML)

            # for returning a list
            ExpressionTable2 <- matrix(0, dim(ExpressionData)[1],dim(ExpressionData)[2] )
            for( i in 1:length(ExpressionData[,1]))
              ExpressionTable2[i,] <- round(ExpressionData[order.pvalue[i],],6)
            rownames(ExpressionTable2) <- GeneId1[modelE$ind.genes[order.pvalue]]
            names <- NULL
            for (i in 1:ncomp2)
              names <- c(names, paste("Loading ",i,sep=""))
            for (i in 1:ncomp2)
              names <- c(names, paste("Correlation ",i,sep=""))
            if (IsContinuousCov)
              colnames(ExpressionTable2) <- c(names, "Weights") else
              colnames(ExpressionTable2) <- names
            resList <- c(resList, list(loadingMatrixModCo=ExpressionTable2))

         } # end of if (length(modelCo$ind.genes) > 0 & !uselasso)

         # Write table of arrays, survival time, censoring indicator and the predicted percentile
         if (any(modelCo$cvpred != "NA") && length(modelCo$ind.genes) > 0) { ## Ting 04/22/15
             if (generateHTML)
             cat("<P></P>", "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>",
                 "<P></P><A name=\"PerctCo\"><B><U>Table of arrays, survival time, censoring indicator and survival risk prediction of training samples.</U></B></A>",
                 file=ExternalFileName, sep="\n", append=T)
             ExpressionData <- data.frame(array = as.character(experimentsTrain),
                                          stime = round(tme[training == 1],3),
                                          status = status[training == 1],
                                          # perct = round(modelCo$cvperct, 1),
                                          pred = as.character(modelCo$cvpred))
             colnames(ExpressionData) <- c("Array", "Survival time", "Censoring indicator",
                                         #"Predicted percentile",
                                         "Predicted risk")
             # for returning a list
             resList <- c(resList, list(predictRiskTrainingModCo=ExpressionData))
             if (generateHTML){
             ExpressionTable <- HTML.data.frame.my(ExpressionData)
             cat("<P></P>", ExpressionTable, "\n", file=ExternalFileName, append= TRUE)
             } # end of if (generateHTML)
         }
     } # End of if (!is.null(clin))

     # Creating time-dependent ROC curve
    if (generateHTML){
     if (file.exists(PlotrocFileName)) {
        cat("<P></P>", "<HR SIZE=1 WIDTH=\"100%\" NOSHADE>",
  	        "<H3><a name=\"rocE\"><B>Time-dependent ROC curve:</B></a></H3>"
             , file=ExternalFileName, sep="\n", append=T)
        if (nperm > 0) {
  	       cat("Permutation p-value for testing the null hypothesis of AUC = 0.5 is ",
  		       format.pval(out$pval.roc[1], eps = 1/nperm), "<BR>\n"
  			   ,file=ExternalFileName, sep="\n", append=T)
  		  }
  	    if (nperm > 0 && !is.na(out$pval.roc[3])) {
              cat("Permutation p-value for testing covariates only model vs combined model is ",
      	       format.pval(out$pval.roc[3], eps=1/nperm), "<BR>\n"
  			   ,file=ExternalFileName, sep="\n", append=T)
  		  }
        cat(paste("<IMG SRC=\"", "rocModelE.png", "\">", "</P>", sep=""),
              file=ExternalFileName, sep="\n", append=T)
     }
     cat("<P></P>", "<HR SIZE=5 WIDTH=\"100%\" NOSHADE>",
  	    "survivalROC package version", packageDescription("survivalROC", fields = "Version"), "</P>", file=ExternalFileName, append=TRUE)
     if (uselasso) {
        cat("glmnet package version", packageDescription("glmnet", fields = "Version"), "</P>", file=ExternalFileName, append=TRUE)
     }
     cat("survriskpred package version", packageDescription("survriskpred", fields = "Version"), "</P>", file=ExternalFileName, append=TRUE)
    } # end of if (generateHTML)
     } # End of if (err == 0)


  if (IsMasterGene) {  # True: create master annotation file, Yong 8/21/06
    # create the gene ID Table
    load(paste(WorkPath, "/GeneIdNames.rda", sep=""))   #use load/save method, Yong 12/6/06
    NumGeneIds <- length(GeneIdNames)
    load(paste(WorkPath, "/GeneId1.rda", sep=""))
    NumGenesInAnalysis <- length(GeneId1)
    GeneId <- matrix("", nrow=NumGenesInAnalysis, ncol=NumGeneIds)
    GeneId[,1] <- GeneId1
    if (NumGeneIds > 1) {
      for (i in 2:NumGeneIds) {
        load(paste(WorkPath, "/GeneId", i, ".rda", sep=""))
        GeneId[,i] <- get(paste("GeneId",i,sep=""))
      }
    }
    # create the master gene IDs <Annotations.html>; info link on the main html will link to this file
    Master.genest <- GeneId[Master.ind,,drop=FALSE]
    CreateMasterAnnotationsTable(projectPath, Master.genest
                                 ,FolderPath=paste(projectPath, "/Output/", analysisName, sep="")
                                 ,AnnotationsTableFile="Annotations.html")
  } # end of if (IsMasterGene = True)
  return(resList)
}


AverageDuplicLGNew <- function(mat, duplic, pheno,training, id=NULL,NumCov=NULL, 
                               Cov1=NULL,Cov2=NULL,Cov3=NULL,Cov4=NULL,Cov5=NULL,Cov6=NULL,
                               RemoveUnpaired = F) {
  # Input:
  #   mat: a matrix
  #   duplic=a string of characters
  #   pheno=a string of characters
  #   id=a string of characters
  # Assume that for rows with the same duplic, pheno and id's
  # also identical

  nc <- length(unique(duplic))
  newmat <- matrix(0, nr=nrow(mat), nc=nc)
  newpheno <- rep(0, nc)
  newduplic <- rep(0, nc)
  newid <- rep(0, nc)
  newtraining<- rep(0,nc)
  newCov1<-NULL
  newCov2<-NULL
  newCov3<-NULL
  newCov4<-NULL
  newCov5<-NULL
  if (!is.null(NumCov))
  {
    if(NumCov>=1)
   {
     newCov1<-rep(0,nc)
    }
   if(NumCov>=2)
   {
   newCov2<-rep(0,nc)
   }
   if(NumCov>=3)
   {
   newCov3<-rep(0,nc)
   }
   if(NumCov>=4)
   {
   newCov4<-rep(0,nc)
   }
   if(NumCov>=5)
   {
   newCov5<-rep(0,nc)
   }
   if(NumCov>=6)
   {
   newCov6<-rep(0,nc)
   }
  }
  j <- 0
  for(i in sort(unique(duplic)))   {
     ind <- which(duplic==i)
     if (length(ind)==1) {
       tmp <- mat[,ind]
     } else {
       tmp <- apply(mat[,ind],1,mean, na.rm=TRUE)
     }
     j <- j+1
     newmat[,j] <- tmp
     newpheno[j] <- pheno[ind[1]]
     newduplic[j] <- duplic[ind[1]]
     newtraining[j]<- training[ind[1]] 
     if(!is.null(id)) {newid[j] <- id[ind[1]]}
     if(!is.null(Cov1)){ newCov1[j]<-Cov1[ind[1]]}
     if(!is.null(Cov2)){ newCov2[j]<-Cov2[ind[1]]}
     if(!is.null(Cov3)){ newCov3[j]<-Cov3[ind[1]]}
     if(!is.null(Cov4)){ newCov4[j]<-Cov4[ind[1]]}
     if(!is.null(Cov5)){ newCov5[j]<-Cov5[ind[1]]}           
	 if(!is.null(Cov6)){ newCov6[j]<-Cov6[ind[1]]}
  }
  if(is.null(id)) {
    dimnames(newmat) <- NULL
    return(list(mat=newmat, duplic=newduplic, pheno=newpheno))  
  } else {
        if(RemoveUnpaired) {
            code <- rep(T,length(newid))
            for(i in 1:length(newid) ) {
                 code[i] <- length(newid[newid == newid[i] ]) == 2
            }
            newmat       <- newmat[,code]
            newid        <- newid[code]
            newduplic    <- newduplic[code]
            newpheno     <- newpheno[code]
            newtraining  <- newtraining[code]
            newCov1<-newCov1[code]
            newCov2<-newCov2[code]    
            newCov3<-newCov3[code]
            newCov4<-newCov4[code]
            newCov5<-newCov5[code]
			newCov6<-newCov6[code]
        }
       
       return(list(mat=newmat, duplic=newduplic, pheno=newpheno,training=newtraining,id=newid,Cov1=newCov1,Cov2=newCov2,Cov3=newCov3,Cov4=newCov4,Cov5=newCov5))
  }  
}

writeFilesRiskPrediction <- function(DoCovariates, covAll, training, StatusVariable, TimeVariable,
                                     KFoldCV, nriskgroups, Percentile1, Percentile2, HasDuplic, arr1, 
                                     nperm, duplic, WorkPath) {
  # Called by createinput_srp
  # Input:
  #   covAll: covariates from training and testing samples
  #   KFoldCV: TRUE/FALSE
  #   HasDuplic: TRUE/FALSE
  #   arr1: gene expression (include predict arrays)
  #   Perm: TRUE/FALSE (deleted) when nperm = 0, permutation tests are OFF.
  #   StatusVariable: should include predict samples
  #   TimeVariable: should include predict samples
  #   ArrayIds: should include predict samples
  #   covAll: should include predict samples
  indic.keep <- rep(TRUE, ncol(arr1))
  ArrayIds <- colnames(arr1)
  #save(list=c("DoCovariates", "NumCov", "StatusVariable","TimeVariable",
  #            "KFoldCV", "Risk2", "HasDuplic", "training", "arr1", 
  #             "Perm", "Covariate1","duplic","indic.keep","ArrayIds"),
  #     file=paste(ArrayToolsPath, "/R/Risktest.rda", sep=""))
  
  NumCov <- ncol(covAll)
  if (DoCovariates) {
    if (NumCov == 1) {
      Covariate1 <- covAll[, 1]
      indic2.keep <- !is.na(Covariate1)
      if (!identical(indic.keep,indic2.keep))  training <- training[tempindex]   
      indic.keep <- indic.keep & indic2.keep  
      Covariate1 <- Covariate1[indic.keep]    
    }
    if (NumCov == 2) {
      Covariate2 <- covAll[, 2]
      indic2.keep <- !is.na(Covariate1) & !is.na(Covariate2) 
      if (!identical(indic.keep,indic2.keep))  training <- training[tempindex]  
      indic.keep <- indic.keep & indic2.keep
      Covariate1 <- Covariate1[indic.keep]
      Covariate2 <- Covariate2[indic.keep] 
    }
    if (NumCov ==3) {
      Covariate3 <- covAll[, 3]
      indic2.keep <- !is.na(Covariate1) & !is.na(Covariate2) & !is.na(Covariate3)
      if (!identical(indic.keep,indic2.keep))  training <- training[tempindex]  
      indic.keep <- indic.keep & indic2.keep
      Covariate1 <- Covariate1[indic.keep]
      Covariate2 <- Covariate2[indic.keep] 
      Covariate3 <- Covariate3[indic.keep]
    }
    if (NumCov ==4) {
      Covariate4 <- covAll[, 4]
      indic2.keep <- !is.na(Covariate1) & !is.na(Covariate2) & !is.na(Covariate3) & !is.na(Covariate4)
      if (!identical(indic.keep,indic2.keep))  training <- training[tempindex]  
      indic.keep <- indic.keep & indic2.keep
      Covariate1 <- Covariate1[indic.keep]
      Covariate2 <- Covariate2[indic.keep] 
      Covariate3 <- Covariate3[indic.keep]
      Covariate4 <- Covariate4[indic.keep]    
    }
    if (NumCov ==5) {
      Covariate5 <- covAll[, 5]
      indic2.keep <- !is.na(Covariate1) & !is.na(Covariate2) & !is.na(Covariate3) & !is.na(Covariate4) & !is.na(Covariate5)
      if (!identical(indic.keep,indic2.keep))  training <- training[tempindex]  
      indic.keep <- indic.keep & indic2.keep
      Covariate1 <- Covariate1[indic.keep]
      Covariate2 <- Covariate2[indic.keep] 
      Covariate3 <- Covariate3[indic.keep]
      Covariate4 <- Covariate4[indic.keep]
      Covariate5 <- Covariate5[indic.keep]
    }
    if (NumCov ==6) {
      Covariate6 <- covAll[, 6]
      indic2.keep <- !is.na(Covariate1) & !is.na(Covariate2) & !is.na(Covariate3) & !is.na(Covariate4) & !is.na(Covariate5) & !is.na(Covariate6)
      if (!identical(indic.keep,indic2.keep))  training <- training[tempindex]  
      indic.keep <- indic.keep & indic2.keep
      Covariate1 <- Covariate1[indic.keep]
      Covariate2 <- Covariate2[indic.keep] 
      Covariate3 <- Covariate3[indic.keep]
      Covariate4 <- Covariate4[indic.keep]
      Covariate5 <- Covariate5[indic.keep]
      Covariate6 <- Covariate6[indic.keep]
    }  
  } # end of if(DoCovariates)
  
  pheno <- paste(StatusVariable[indic.keep],TimeVariable[indic.keep],sep="ZZZ")
  phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
  options(warn=-1)
  phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
  options(warn=0)
  
  arr1 <- arr1[,indic.keep, drop = FALSE] ####To subset the excluded arrays if any 02/15/2005
  gc() # added Ming-Chung 11/28/06
  
  if (HasDuplic) {
    if(DoCovariates) {
       if(NumCov==1) {
         AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds,NumCov,Covariate1)
         arr1 <- AvgDat$mat
         pheno <- AvgDat$pheno
         duplic <- AvgDat$duplic
         ArrayIds1 <- AvgDat$id
         phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
         options(warn=-1)
         phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
         options(warn=0)
         training<- AvgDat$training
         Covariate1<-AvgDat$Cov1
         rm(AvgDat); gc()
       }
       if(NumCov==2) {
          AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds,NumCov,Covariate1,Covariate2)
          arr1 <- AvgDat$mat
          pheno <- AvgDat$pheno
          duplic <- AvgDat$duplic
          ArrayIds1 <- AvgDat$id
          phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
          options(warn=-1)
          phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
          options(warn=0)
          training<- AvgDat$training
          Covariate1<-AvgDat$Cov1
          Covariate2<-AvgDat$Cov2
          rm(AvgDat); gc()
       }
       if(NumCov==3) {
          AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds,NumCov,Covariate1,Covariate2,Covariate3)
          arr1 <- AvgDat$mat
          pheno <- AvgDat$pheno
          duplic <- AvgDat$duplic
          ArrayIds1 <- AvgDat$id
          phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
          options(warn=-1)
          phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
          options(warn=0)
          training<- AvgDat$training
          Covariate1<-AvgDat$Cov1
          Covariate2<-AvgDat$Cov2
          Covariate3<-AvgDat$Cov3
          rm(AvgDat); gc()
       }
       if(NumCov==4) {
          AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds,NumCov,Covariate1,Covariate2,Covariate3,Covariate4)
          arr1 <- AvgDat$mat
          pheno <- AvgDat$pheno
          duplic <- AvgDat$duplic
          ArrayIds1 <- AvgDat$id
          phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
          options(warn=-1)
          phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
          options(warn=0)
          training<- AvgDat$training
          Covariate1<-AvgDat$Cov1
          Covariate2<-AvgDat$Cov2
          Covariate3<-AvgDat$Cov3
          Covariate4<-AvgDat$Cov4
          rm(AvgDat); gc()
       }
       if(NumCov==5) {
          AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds,NumCov,Covariate1,Covariate2,Covariate3,Covariate4,Covariate5)
          arr1 <- AvgDat$mat
          pheno <- AvgDat$pheno
          duplic <- AvgDat$duplic
          ArrayIds1 <- AvgDat$id
          phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
          options(warn=-1)
          phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
          options(warn=0)
          training<- AvgDat$training
          Covariate1<-AvgDat$Cov1
          Covariate2<-AvgDat$Cov2
          Covariate3<-AvgDat$Cov3
          Covariate4<-AvgDat$Cov4
          Covariate5<-AvgDat$Cov5
          rm(AvgDat); gc()
       }   
       if(NumCov==6) {
          AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds,NumCov,Covariate1,Covariate2,Covariate3,Covariate4,Covariate5)
          arr1 <- AvgDat$mat
          pheno <- AvgDat$pheno
          duplic <- AvgDat$duplic
          ArrayIds1 <- AvgDat$id
          phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
          options(warn=-1)
          phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
          options(warn=0)
          training<- AvgDat$training
          Covariate1<-AvgDat$Cov1
          Covariate2<-AvgDat$Cov2
          Covariate3<-AvgDat$Cov3
          Covariate4<-AvgDat$Cov4
          Covariate5<-AvgDat$Cov5
        	Covariate6<-AvgDat$Cov6
          rm(AvgDat); gc()
       }
    } else { 
      AvgDat <- AverageDuplicLGNew(arr1, duplic,pheno,training,ArrayIds)
      arr1 <- AvgDat$mat
      pheno <- AvgDat$pheno
      duplic <- AvgDat$duplic
      ArrayIds1 <- AvgDat$id
      phenoClass <- as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,1]))
      options(warn=-1)
      phenoTime <- as.double(as.character((t(data.frame(strsplit(pheno,"ZZZ")))[,2])))
      options(warn=0)
      training<- AvgDat$training
      rm(AvgDat); gc()
    } 
  }
  
  if (HasDuplic) {
    ArrayIds <- as.matrix(ArrayIds)
    ind <- ArrayIds1%in%ArrayIds
  }
  ifelse(HasDuplic, experiments <- data.frame(ExpId=duplic)[ind,], 
                    experiments <- data.frame(ExpId=ArrayIds)[indic.keep,]) ### To maintain the same order 02/15/2006
  save(experiments, file=paste(WorkPath,"/experiments.rda", sep=""))
  
  nfold<-10
  nfold<-ifelse(KFoldCV, 10, length(experiments[training[training==1]]))
#  nperm<-ifelse(Perm, nperm, 0)
  options(warn=-1)
  status<-as.numeric(phenoClass)
  options(warn=0)
  tme<-as.numeric(phenoTime)
  perc<- ifelse(nriskgroups == 3, c(Percentile1,Percentile2), Percentile1)
  
  if(DoCovariates)
  {
    options(warn=-1)
    if(NumCov >= 1) {
      mytmp <- Covariate1[1]
      # convert the covariate to numerical if it is.
      clin <- if (!is.na(as.double(mytmp)) && mytmp == as.character(as.double(mytmp))) {
                 data.frame(cov1 = as.numeric(Covariate1))
              } else data.frame(cov1 = Covariate1)
    } 
    if(NumCov >= 2) {
      mytmp <- Covariate2[1]
      clin.tmp <- if (!is.na(as.double(mytmp)) && mytmp == as.character(as.double(mytmp))) {
                 data.frame(cov2 = as.numeric(Covariate2))
              } else data.frame(cov2 = Covariate2)
      clin <- cbind(clin, clin.tmp)
    }
    if(NumCov >= 3) { 
      mytmp <- Covariate3[1]
      clin.tmp <- if (!is.na(as.double(mytmp)) && mytmp == as.character(as.double(mytmp))) {
                 data.frame(cov3 = as.numeric(Covariate3))
              } else data.frame(cov3 = Covariate3)
      clin <- cbind(clin, clin.tmp)
    }    
    if(NumCov >= 4) { 
      mytmp <- Covariate4[1]
      clin.tmp <- if (!is.na(as.double(mytmp)) && mytmp == as.character(as.double(mytmp))) {
                 data.frame(cov4 = as.numeric(Covariate4))
              } else data.frame(cov4 = Covariate4)
      clin <- cbind(clin, clin.tmp)
    }    
    if(NumCov >= 5) { 
      mytmp <- Covariate5[1]
      clin.tmp <- if (!is.na(as.double(mytmp)) && mytmp == as.character(as.double(mytmp))) {
                 data.frame(cov5 = as.numeric(Covariate5))
              } else data.frame(cov5 = Covariate5)
      clin <- cbind(clin, clin.tmp)
    }    
    if(NumCov >= 6) { 
      mytmp <- Covariate6[1]
      clin.tmp <- if (!is.na(as.double(mytmp)) && mytmp == as.character(as.double(mytmp))) {
                 data.frame(cov6 = as.numeric(Covariate6))
              } else data.frame(cov6 = Covariate6)
      clin <- cbind(clin, clin.tmp)
    }     
    options(warn=0)
  } else {
    clin<-NULL
  }
  save(list=c("status","tme","training", "arr1","nperm","nfold", "clin","perc"),
       file=paste(WorkPath, "/Riskdata.rda", sep="")) # skip 'alpha', 'ncomp', 'uselasso', 'mixing', 'pcrgenes', 'landmarktime'
                                                      # since they are not processed and can be obtained from survriskpredict().
} # end of writefilesriskprediction()
