% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survRiskPredict.R
\name{survRiskPredict}
\alias{survRiskPredict}
\title{Survival risk prediction}
\usage{
survRiskPredict(exprTrain, covTrain, exprTest = NULL, covTest = NULL,
  geneid, status, tme, geneSelect = "pc", nriskgroups = 2,
  progIndexPerc, cvMethod = "10fold", nperm = 0, landmarktime = 0,
  alpha = 0.001, ncomp = 2, mixing = 1, pcrgenes = 10, projectPath,
  outputName = "SurvivalRiskPrediction", generateHTML = FALSE)
}
\arguments{
\item{exprTrain}{matrix/data frame of gene expression data for training samples. Rows are genes and columns are arrays. Its column names must be provided.}

\item{covTrain}{matrix/data frame of clinical covariates from the training samples. Rows are arrays and columns are clinical covariates. Default is NULL.}

\item{exprTest}{matrix/data frame of gene expression data for new samples. Rows are genes and columns are arrays. Its column names must be provided.}

\item{geneid}{matrix/data frame of gene IDs.}

\item{status}{vector specifying survival status (1 = death, 0 = censored).}

\item{tme}{vector specifying survival time.}

\item{geneSelect}{character string of a gene selection method: "pc" for principal component and "pcr" for penalized Cox regression. Default is "pc".}

\item{nriskgroups}{numeric of the number of risk groups. It can only be 2 or 3. Default is 2.}

\item{progIndexPerc}{numeric of prognostic index percentile. In 2-risk group case, one percentile values needs to be provided, whereas, in
the 3-risk group case, two percentile values need to be provided. Default is 50 for 2-risk group case.}

\item{cvMethod}{character string of cross validation method. "loocv" for leave-one-out cross validation, and "10fold" for 10-fold cross validation.}

\item{nperm}{numeric of number of permutation tests. If \code{nperm == 0}, no permutation tests will be conducted; otherwise, permutation tests based
on log rank statistics will be run. Default is 0.}

\item{landmarktime}{numeric of landmark time for the ROC curve. If \code{landmarktime=0}, median of survival time will be used.}

\item{alpha}{numeric of the threshold significance level of penalized Cox regression model for selecting the genes to
be used in computing the principal components. Default is 0.001.}

\item{ncomp}{numeric of number of principal components if \code{geneSelect = "pc"}.}

\item{mixing}{numeric of the mixing parameter in penalized Cox regression.}

\item{pcrgenes}{numeric of number of desired genes if \code{geneSelect = "pcr"}.}

\item{projectPath}{character string specifying the full project path.}

\item{outputName}{character string specifying the output folder name. Default is "SurvivalRiskPrediction".}

\item{generateHTML}{logical. If \code{TRUE}, an HTML page will be generated with detailed class prediction results
saved in <projectPath>/Output/<outputName>/<outputName>.html; otherwise not.}

\item{CovTest}{matrix/data frame of clinical covariates from new samples. Rows are arrays and columns are clinical covariates. Default is NULL.}
}
\value{
A list that may include the following objects:
\itemize{
   \item \code{predictNewSamplesModE}: a data frame with risk group predictions of new samples when the model with gene expression only is considered.
   \item \code{genesInClassifierModE}: a data frame with genes selected by fitting penalized Cox proportional hazards model
                                       when the model with gene expression only is considered.
   \item \code{loadingMatrixModE}: a data frame with loading matrix of the significant genes and the correlations between the principal components
                                   and the significant genes when the model with gene expression only is considered.
   \item \code{predictRiskTrainingModE}: a data frame with Table of arrays, survival time, censoring indicator and survival risk prediction
                                         of training samples when the model with gene expression only is considered.
   \item \code{predictNewSamplesModCo}: a data frame with risk group predictions of new samples when the model with both covariates and gene expression is considered.
   \item \code{genesInClassifierModCo}: a data frame with genes selected by fitting penalized Cox proportional hazards model
                                       when the model with both covariates and gene expression is considered.
   \item \code{loadingMatrixModCo}: a data frame with loading matrix of the significant genes and the correlations between the principal components
                                   and the significant genes when the model with both covariates and gene expression is considered.
   \item \code{predictRiskTrainingModCo}: a data frame with Table of arrays, survival time, censoring indicator and survival risk prediction
                                         of training samples when the model with both covariates and gene expression is considered.
   \item \code{predictNewSamplesModCl}: a data frame with risk group predictions of new samples when the model with covariates only is considered.
   \item \code{predictRiskTrainingModCl}: a data frame with Table of arrays, survival time, censoring indicator and survival risk prediction
                                         of training samples when the model with covariates only is considered.
   \item \code{ppval}: a vector with permutation p-values for the models with gene expression only, covariates only or the combination of gene expression and
                       covariates.
   \item \code{workPath}: the path for Fortran and other intermediate outputs.
   }
}
\description{
This function implements the survial risk prediction tool in BRB-ArrayTools
}
\details{
Please see the BRB-ArrayTools manual for details.
}
\examples{
#example 1
resList1 <- test.survRiskPredict("Pomeroy")
#example2
data("Pomeroy")
projectPath <- tempdir()
outputName <- "SurvivalRiskPrediction"
generateHTML <- TRUE
resList2 <- survRiskPredict(exprTrain, covTrain, exprTest, covTest,
                           geneId, status, tme, geneSelect = "pc",
                           nriskgroups = 2, progIndexPerc = 50,
                           cvMethod = "10fold", nperm = 0, landmarktime = 0,
                           alpha = .001, ncomp = 2, mixing = 1, pcrgenes = 10,
                           projectPath = projectPath,
                           outputName = outputName, generateHTML)
if (generateHTML)
  browseURL(file.path(projectPath, "Output", outputName,
            paste0(outputName, ".html")))
#example3
data("Perou")
projectPath <- file.path(Sys.getenv("HOME"),"Pomeroy")
outputName <- "SurvivalRiskPrediction"
generateHTML <- TRUE
resList3 <- survRiskPredict(exprTrain, covTrain, exprTest, covTest,
                           geneId, status, tme, geneSelect = "pc",
                           nriskgroups = 2, progIndexPerc = 50,
                           cvMethod = "10fold", nperm = 0, landmarktime = 0,
                           alpha = .001, ncomp = 2, mixing = 1, pcrgenes = 10,
                           projectPath = projectPath,
                           outputName = outputName, generateHTML)
if (generateHTML)
  browseURL(file.path(projectPath, "Output", outputName,
            paste0(outputName, ".html")))
}
